from pdb import set_trace as T
import numpy as np

import torch
from torch import nn


def modelSize(net):
   '''Print model size'''
   params = 0
   for e in net.parameters():
      params += np.prod(e.size())
   params = int(params/1000)
   print("Network has ", params, "K params")

def ModuleList(module, *args, n=1):
   '''Repeat module n times'''
   return nn.ModuleList([module(*args) for i in range(n)])

#Full-network initialization wrapper
def initWeights(net, scheme='orthogonal'):
   '''Provides multiple weight initialization schemes

   Args:
      net: network to initialize
      scheme: otrhogonal, normal, or xavier
   '''
   print('Initializing weights. Warning: may overwrite sensitive bias parameters (e.g. batchnorm)')
   for e in net.parameters():
      if scheme == 'orthogonal':
         if len(e.size()) >= 2:
            init.orthogonal(e)
      elif scheme == 'normal':
         init.normal(e, std=1e-2)
      elif scheme == 'xavier':
         init.xavier_normal(e)

def pack(val):
   '''Pack values to tensor'''
   seq_lens   = torch.LongTensor(list(map(len, val)))
   seq_tensor = torch.zeros((len(val), seq_lens.max()))
   for idx, (seq, seqlen) in enumerate(zip(val, seq_lens)):
      seq_tensor[idx, :seqlen] = torch.Tensor(seq)

   #Todo: reintroduce sort
   #seq_lens, perm_idx = seq_lens.sort(0, descending=True)
   #seq_tensor = seq_tensor[perm_idx]

   return seq_tensor, seq_lens

#Be sure to unsort these
def unpack(vals, lens):
   '''Unpack value tensor using provided lens'''
   ret = []
   for idx, l in enumerate(lens):
      e = vals[idx, :l]
      ret.append(e)
   return ret




#from .droptable import DropTable
#from .exchange import Exchange
#from .recipe import Recipe
#from .inventory import Inventory
from neural_mmo.forge.blade.systems.skill import Skill
from neural_mmo.forge.blade.systems.visualizer import visualizer



from pdb import set_trace as T
import numpy as np
import torch

def zeroGrads(ann):
   '''Zeros out gradients

   Args:
      ann: a model
   '''
   ind = 0
   for e in ann.parameters():
      if e.grad is None:
          continue
      shape = e.size()
      nParams = np.prod(shape)
      e.grad.data *= 0
      ind += nParams

def setParameters(ann, meanVec):
   '''Sets the parameters of the ann

   Args:
      meanVec: A list of parameters
   '''
   if meanVec is None:
      return

   ind = 0
   for e in ann.parameters():
      shape = e.size()
      nParams = np.prod(shape)
      data = np.array(meanVec[ind:ind+nParams]).reshape(*shape)
      e.data = torch.Tensor(data).to(e.device)
      ind += nParams

def setGrads(ann, grads):
   '''Sets the gradients of the ann
      
   Args:
      grads: A list of gradients
   '''
   ind = 0
   for e in ann.parameters():
      shape = e.size()
      nParams = np.prod(shape)
      e.grad.data = torch.Tensor(np.array(grads[ind:ind+nParams]).reshape(*shape))
      ind += nParams

def getParameters(ann):
   '''Get the parameters of the ann

   Args:
      ann: The model to get parameters from

   Returns:
      ret: A list of parameters
   '''
   ret = []
   for e in ann.parameters():
      ret += e.data.cpu().view(-1).numpy().tolist()
   return ret

def getGrads(ann, warn=True):
   '''Get the gradients of the ann

   Args:
      ann: The model to get gradients from
      warn (bool): Whether to warn when gradients are None.
         This is a common red flag that there is something
         wrong with the network or training

   Returns:
      ret: A list of gradients
   '''
   ret = []
   for param, e in ann.named_parameters():
      if e.grad is None:
         if warn:
            print(str(param), ': GRADIENT NOT FOUND. Possible causes: (1) you have loaded a model with a different architecture. (2) This layer is not differentiable or not in use.')
         ret += np.zeros(e.shape).ravel().tolist()
      else:
         #nan = torch.sum(e.grad != e.grad) > 0:
         dat = e.grad.data.cpu().view(-1).numpy().tolist()
         assert sum(np.isnan(dat)) == 0
         ret += dat
   return ret




from pdb import set_trace as T
import numpy as np

import torch
from torch import nn
from torch.distributions import Categorical

def classify(logits, mask=None, eps=1e-3):
   '''Sample an action from logits'''
   if len(logits.shape) == 1:
      logits = logits.view(1, -1)

   #Masking the noise required. Do not do this
   #in place. We do not want to -inf pad logits
   #for pg loss, only for Categorical sampling
   logits = logits + eps
   if mask is not None:
      logits[~mask] = -np.inf

   distribution = Categorical(logits=logits)
   atn = distribution.sample()
   return atn

def dot(k, v):
   '''Compute v*kT (transpose on last two dims)'''
   kt = k.transpose(-1, -2)
   x = torch.matmul(v, kt)
   x = x.squeeze(-1)
   return x




from collections import defaultdict

from tqdm import tqdm
import numpy as np
import gym
import wandb
import trueskill

import torch
from torch import nn
from torch.nn.utils import rnn

from ray import rllib
from pdb import set_trace as TT

import ray.rllib.agents.ppo.ppo as ppo
import ray.rllib.agents.ppo.appo as appo
import ray.rllib.agents.impala.impala as impala
from ray.rllib.agents.callbacks import DefaultCallbacks
from ray.tune.integration.wandb import WandbLoggerCallback
from ray.rllib.utils.spaces.flexdict import FlexDict
from ray.rllib.models.torch.recurrent_net import RecurrentNetwork

from neural_mmo.forge.blade.io.action.static import Action, Fixed
from neural_mmo.forge.blade.io.stimulus.static import Stimulus
from neural_mmo.forge.blade.lib import overlay
from neural_mmo.forge.blade.systems import ai

from neural_mmo.forge.ethyr.torch import policy
from neural_mmo.forge.ethyr.torch.policy import attention

from neural_mmo.forge.trinity import Env
from neural_mmo.forge.trinity.dataframe import DataType
from neural_mmo.forge.trinity.overlay import Overlay, OverlayRegistry


###############################################################################
### Pytorch model + IO. The pip package contains some submodules
class Input(nn.Module):
   def __init__(self, config, embeddings, attributes):
      '''Network responsible for processing observations

      Args:
         config     : A configuration object
         embeddings : An attribute embedding module
         attributes : An attribute attention module
      '''
      super().__init__()

      self.embeddings = nn.ModuleDict()
      self.attributes = nn.ModuleDict()

      for _, entity in Stimulus:
         continuous = len([e for e in entity if e[1].CONTINUOUS])
         discrete   = len([e for e in entity if e[1].DISCRETE])
         self.attributes[entity.__name__] = nn.Linear(
               (continuous+discrete)*config.HIDDEN, config.HIDDEN)
         self.embeddings[entity.__name__] = embeddings(
               continuous=continuous, discrete=4096, config=config)

      #Hackey obs scaling
      self.tileWeight = torch.Tensor([1.0, 0.0, 0.02, 0.02])
      self.entWeight  = torch.Tensor([1.0, 0.0, 0.0, 0.05, 0.00, 0.02, 0.02, 0.1, 0.01, 0.1, 0.1, 0.1, 0.3])

   def forward(self, inp):
      '''Produces tensor representations from an IO object

      Args:                                                                   
         inp: An IO object specifying observations                      
         
      Returns:
         entityLookup: A fixed size representation of each entity
      ''' 
      #Pack entities of each attribute set
      entityLookup = {}

      device                       = inp['Tile']['Continuous'].device
      inp['Tile']['Continuous']   *= self.tileWeight.to(device)
      inp['Entity']['Continuous'] *= self.entWeight.to(device)
 
      entityLookup['N'] = inp['Entity'].pop('N')
      for name, entities in inp.items():
         #Construct: Batch, ents, nattrs, hidden
         embeddings = self.embeddings[name](entities)
         B, N, _, _ = embeddings.shape
         embeddings = embeddings.view(B, N, -1)

         #Construct: Batch, ents, hidden
         entityLookup[name] = self.attributes[name](embeddings)

      return entityLookup

class Output(nn.Module):
   def __init__(self, config):
      '''Network responsible for selecting actions

      Args:
         config: A Config object
      '''
      super().__init__()
      self.config = config
      self.h = config.HIDDEN

      self.net = DiscreteAction(self.config, self.h, self.h)
      self.arg = nn.Embedding(Action.n, self.h)

   def names(self, nameMap, args):
      '''Lookup argument indices from name mapping'''
      return np.array([nameMap.get(e) for e in args])

   def forward(self, obs, lookup):
      '''Populates an IO object with actions in-place                         
                                                                              
      Args:                                                                   
         obs     : An IO object specifying observations
         lookup  : A fixed size representation of each entity
      ''' 
      rets = defaultdict(dict)
      for atn in Action.edges:
         for arg in atn.edges:
            lens  = None
            if arg.argType == Fixed:
               batch = obs.shape[0]
               idxs  = [e.idx for e in arg.edges]
               cands = self.arg.weight[idxs]
               cands = cands.repeat(batch, 1, 1)
            else:
               cands = lookup['Entity']
               lens  = lookup['N']

            logits = self.net(obs, cands, lens)
            rets[atn][arg] = logits

      return rets
      
class DiscreteAction(nn.Module):
   '''Head for making a discrete selection from
   a variable number of candidate actions'''
   def __init__(self, config, xdim, h):
      super().__init__()
      self.net = attention.DotReluBlock(h)

   def forward(self, stim, args, lens):
      x = self.net(stim, args)

      if lens is not None:
         mask = torch.arange(x.shape[-1]).to(x.device).expand_as(x)
         x[mask >= lens] = 0

      return x

class Base(nn.Module):
   def __init__(self, config):
      '''Base class for baseline policies

      Args:
         config: A Configuration object
      '''
      super().__init__()
      self.embed  = config.EMBED
      self.config = config

      self.output = Output(config)
      self.input  = Input(config,
            embeddings=policy.MixedDTypeInput,
            attributes=policy.SelfAttention)

      self.valueF = nn.Linear(config.HIDDEN, 1)

   def hidden(self, obs, state=None, lens=None):
      '''Abstract method for hidden state processing, recurrent or otherwise,
      applied between the input and output modules

      Args:
         obs: An observation dictionary, provided by forward()
         state: The previous hidden state, only provided for recurrent nets
         lens: Trajectory segment lengths used to unflatten batched obs
      '''
      raise NotImplementedError('Implement this method in a subclass')

   def forward(self, obs, state=None, lens=None):
      '''Applies builtin IO and value function with user-defined hidden
      state subnetwork processing. Arguments are supplied by RLlib
      '''
      entityLookup  = self.input(obs)
      hidden, state = self.hidden(entityLookup, state, lens)
      self.value    = self.valueF(hidden).squeeze(1)
      actions       = self.output(hidden, entityLookup)
      return actions, state

class Encoder(Base):
   def __init__(self, config):
      '''Simple baseline model with flat subnetworks'''
      super().__init__(config)
      h = config.HIDDEN

      self.ent    = nn.Linear(2*h, h)
      self.conv   = nn.Conv2d(h, h, 3)
      self.pool   = nn.MaxPool2d(2)
      self.fc     = nn.Linear(h*6*6, h)

      self.proj   = nn.Linear(2*h, h)
      self.attend = policy.SelfAttention(self.embed, h)

   def hidden(self, obs, state=None, lens=None):
      #Attentional agent embedding
      agentEmb  = obs['Entity']
      selfEmb   = agentEmb[:, 0:1].expand_as(agentEmb)
      agents    = torch.cat((selfEmb, agentEmb), dim=-1)
      agents    = self.ent(agents)
      agents, _ = self.attend(agents)
      #agents = self.ent(selfEmb)

      #Convolutional tile embedding
      tiles     = obs['Tile']
      self.attn = torch.norm(tiles, p=2, dim=-1)

      w      = self.config.WINDOW
      batch  = tiles.size(0)
      hidden = tiles.size(2)
      #Dims correct?
      tiles  = tiles.reshape(batch, w, w, hidden).permute(0, 3, 1, 2)
      tiles  = self.conv(tiles)
      tiles  = self.pool(tiles)
      tiles  = tiles.reshape(batch, -1)
      tiles  = self.fc(tiles)

      hidden = torch.cat((agents, tiles), dim=-1)
      hidden = self.proj(hidden)
      return hidden, state

class Recurrent(Encoder):
   def __init__(self, config):
      '''Recurrent baseline model'''
      super().__init__(config)
      self.lstm   = policy.BatchFirstLSTM(
            input_size=config.HIDDEN,
            hidden_size=config.HIDDEN)

   #Note: seemingly redundant transposes are required to convert between 
   #Pytorch (seq_len, batch, hidden) <-> RLlib (batch, seq_len, hidden)
   def hidden(self, obs, state, lens):
      #Attentional input preprocessor and batching
      lens = lens.cpu() if type(lens) == torch.Tensor else lens
      hidden, _ = super().hidden(obs)
      config    = self.config
      h, c      = state

      TB = hidden.size(0) #Padded batch of size (seq x batch)
      B  = len(lens)      #Sequence fragment time length
      T  = TB // B        #Trajectory batch size
      H  = config.HIDDEN  #Hidden state size

      #Pack (batch x seq, hidden) -> (batch, seq, hidden)
      hidden        = rnn.pack_padded_sequence(
                         input=hidden.view(B, T, H),
                         lengths=lens,
                         enforce_sorted=False,
                         batch_first=True)

      #Main recurrent network
      oldHidden = hidden
      hidden, state = self.lstm(hidden, state)
      newHidden = hidden

      #Unpack (batch, seq, hidden) -> (batch x seq, hidden)
      hidden, _     = rnn.pad_packed_sequence(
                         sequence=hidden,
                         batch_first=True,
                         total_length=T)

      return hidden.reshape(TB, H), state

###############################################################################
### RLlib Policy, Evaluator, Trainer
class RLlibPolicy(RecurrentNetwork, nn.Module):
   '''Wrapper class for using our baseline models with RLlib'''
   def __init__(self, *args, **kwargs):
      self.config = kwargs.pop('config')
      super().__init__(*args, **kwargs)
      nn.Module.__init__(self)

      self.space  = actionSpace(self.config).spaces
      self.model  = Recurrent(self.config)

   #Initial hidden state for RLlib Trainer
   def get_initial_state(self):
      return [self.model.valueF.weight.new(1, self.config.HIDDEN).zero_(),
              self.model.valueF.weight.new(1, self.config.HIDDEN).zero_()]

   def forward(self, input_dict, state, seq_lens):
      logitDict, state = self.model(input_dict['obs'], state, seq_lens)

      logits = []
      #Flatten structured logits for RLlib
      for atnKey, atn in sorted(self.space.items()):
         for argKey, arg in sorted(atn.spaces.items()):
            logits.append(logitDict[atnKey][argKey])

      return torch.cat(logits, dim=1), state

   def value_function(self):
      return self.model.value

   def attention(self):
      return self.model.attn


###############################################################################
### RLlib Wrappers: Env, Overlays
class RLlibEnv(Env, rllib.MultiAgentEnv):
   def __init__(self, config):
      self.config = config['config']
      super().__init__(self.config)

   def reward(self, ent):
      config      = self.config

      ACHIEVEMENT = config.REWARD_ACHIEVEMENT
      SCALE       = config.ACHIEVEMENT_SCALE
      COOPERATIVE = config.COOPERATIVE

      individual  = 0 if ent.entID in self.realm.players else -1
      team        = 0

      if ACHIEVEMENT:
         individual += SCALE*ent.achievements.update(self.realm, ent, dry=True)
      if COOPERATIVE:
         nDead = len([p for p in self.dead.values() if p.population == ent.pop])
         team  = -nDead / config.TEAM_SIZE
      if COOPERATIVE and ACHIEVEMENT:
         pre, post = [], []
         for p in self.realm.players.corporeal.values():
            if p.population == ent.pop:
               pre.append(p.achievements.score(aggregate=False))
               post.append(p.achievements.update(
                     self.realm, ent, aggregate=False, dry=True))
        
         pre   = np.array(pre).max(0)
         post  = np.array(post).max(0)
         team += SCALE*(post - pre).sum()

      ent.achievements.update(self.realm, ent)

      alpha  = config.TEAM_SPIRIT
      return alpha*team + (1.0-alpha)*individual

   def step(self, decisions, preprocess=None, omitDead=False):
      preprocess = {entID for entID in decisions}
      obs, rewards, dones, infos = super().step(decisions, preprocess, omitDead)

      config = self.config
      dones['__all__'] = False
      test = config.EVALUATE or config.RENDER
      
      if config.EVALUATE:
         horizon = config.EVALUATION_HORIZON
      else:
         horizon = config.TRAIN_HORIZON

      population  = len(self.realm.players) == 0
      hit_horizon = self.realm.tick >= horizon
      
      if not config.RENDER and (hit_horizon or population):
         dones['__all__'] = True

      return obs, rewards, dones, infos

def observationSpace(config):
   obs = FlexDict(defaultdict(FlexDict))
   for entity in sorted(Stimulus.values()):
      nRows       = entity.N(config)
      nContinuous = 0
      nDiscrete   = 0

      for _, attr in entity:
         if attr.DISCRETE:
            nDiscrete += 1
         if attr.CONTINUOUS:
            nContinuous += 1

      obs[entity.__name__]['Continuous'] = gym.spaces.Box(
            low=-2**20, high=2**20, shape=(nRows, nContinuous),
            dtype=DataType.CONTINUOUS)

      obs[entity.__name__]['Discrete']   = gym.spaces.Box(
            low=0, high=4096, shape=(nRows, nDiscrete),
            dtype=DataType.DISCRETE)

   obs['Entity']['N']   = gym.spaces.Box(
         low=0, high=config.N_AGENT_OBS, shape=(1,),
         dtype=DataType.DISCRETE)
   return obs

def actionSpace(config):
   atns = FlexDict(defaultdict(FlexDict))
   for atn in sorted(Action.edges):
      for arg in sorted(atn.edges):
         n              = arg.N(config)
         atns[atn][arg] = gym.spaces.Discrete(n)
   return atns

class RLlibOverlayRegistry(OverlayRegistry):
   '''Host class for RLlib Map overlays'''
   def __init__(self, realm):
      super().__init__(realm.config, realm)

      self.overlays['values']       = Values
      self.overlays['attention']    = Attention
      self.overlays['tileValues']   = TileValues
      self.overlays['entityValues'] = EntityValues

class RLlibOverlay(Overlay):
   '''RLlib Map overlay wrapper'''
   def __init__(self, config, realm, trainer, model):
      super().__init__(config, realm)
      self.trainer = trainer
      self.model   = model

class Attention(RLlibOverlay):
   def register(self, obs):
      '''Computes local attentional maps with respect to each agent'''
      tiles      = self.realm.realm.map.tiles
      players    = self.realm.realm.players

      attentions = defaultdict(list)
      for idx, playerID in enumerate(obs):
         if playerID not in players:
            continue
         player = players[playerID]
         r, c   = player.pos

         rad     = self.config.NSTIM
         obTiles = self.realm.realm.map.tiles[r-rad:r+rad+1, c-rad:c+rad+1].ravel()

         for tile, a in zip(obTiles, self.model.attention()[idx]):
            attentions[tile].append(float(a))

      sz    = self.config.TERRAIN_SIZE
      data  = np.zeros((sz, sz))
      for r, tList in enumerate(tiles):
         for c, tile in enumerate(tList):
            if tile not in attentions:
               continue
            data[r, c] = np.mean(attentions[tile])

      colorized = overlay.twoTone(data)
      self.realm.register(colorized)

class Values(RLlibOverlay):
   def update(self, obs):
      '''Computes a local value function by painting tiles as agents
      walk over them. This is fast and does not require additional
      network forward passes'''
      players = self.realm.realm.players
      for idx, playerID in enumerate(obs):
         if playerID not in players:
            continue
         r, c = players[playerID].base.pos
         self.values[r, c] = float(self.model.value_function()[idx])

   def register(self, obs):
      colorized = overlay.twoTone(self.values[:, :])
      self.realm.register(colorized)

def zeroOb(ob, key):
   for k in ob[key]:
      ob[key][k] *= 0

class GlobalValues(RLlibOverlay):
   '''Abstract base for global value functions'''
   def init(self, zeroKey):
      if self.trainer is None:
         return

      print('Computing value map...')
      model     = self.trainer.get_policy('policy_0').model
      obs, ents = self.realm.dense()
      values    = 0 * self.values

      #Compute actions to populate model value function
      BATCH_SIZE = 128
      batch = {}
      final = list(obs.keys())[-1]
      for agentID in tqdm(obs):
         ob             = obs[agentID]
         batch[agentID] = ob
         zeroOb(ob, zeroKey)
         if len(batch) == BATCH_SIZE or agentID == final:
            self.trainer.compute_actions(batch, state={}, policy_id='policy_0')
            for idx, agentID in enumerate(batch):
               r, c         = ents[agentID].base.pos
               values[r, c] = float(self.model.value_function()[idx])
            batch = {}

      print('Value map computed')
      self.colorized = overlay.twoTone(values)

   def register(self, obs):
      print('Computing Global Values. This requires one NN pass per tile')
      self.init()

      self.realm.register(self.colorized)

class TileValues(GlobalValues):
   def init(self, zeroKey='Entity'):
      '''Compute a global value function map excluding other agents. This
      requires a forward pass for every tile and will be slow on large maps'''
      super().init(zeroKey)

class EntityValues(GlobalValues):
   def init(self, zeroKey='Tile'):
      '''Compute a global value function map excluding tiles. This
      requires a forward pass for every tile and will be slow on large maps'''
      super().init(zeroKey)


class RLlibTrainer(ppo.PPOTrainer):
   def __init__(self, config, env=None, logger_creator=None):
      super().__init__(config, env, logger_creator)
      self.env_config = config['env_config']['config']

      #1/sqrt(2)=76% win chance within beta, 95% win chance vs 3*beta=100 SR
      trueskill.setup(mu=1000, sigma=2*100/3, beta=100/3, tau=2/3, draw_probability=0)

      self.ratings = [{agent.__name__: trueskill.Rating(mu=1000, sigma=2*100/3)}
            for agent in self.env_config.EVAL_AGENTS]

      self.reset_scripted()

   def reset_scripted(self):
      for rating_dict in self.ratings:
         for agent, rating in rating_dict.items():
            if agent == 'Combat':
               rating_dict[agent] = trueskill.Rating(mu=1500, sigma=1)

   def post_mean(self, stats):
      for key, vals in stats.items():
          if type(vals) == list:
              stats[key] = np.mean(vals)

   def train(self):
      stats = super().train()
      self.post_mean(stats['custom_metrics'])
      return stats

   def evaluate(self):
      stat_dict = super().evaluate()
      stats = stat_dict['evaluation']['custom_metrics']
 
      ranks = {agent.__name__: -1 for agent in self.env_config.EVAL_AGENTS}
      for key in list(stats.keys()):
         if key.startswith('Rank_'):
             stat = stats[key]
             del stats[key]
             agent = key[5:]
             ranks[agent] = stat

      ranks = list(ranks.values())
      nEnvs = len(ranks[0])
      
      #Once RLlib adds better custom metric support,
      #there should be a cleaner way to divide episodes into blocks
      for i in range(nEnvs): 
         env_ranks = [e[i] for e in ranks]
         self.ratings = trueskill.rate(self.ratings, env_ranks)
         self.reset_scripted()

      for rating in self.ratings:
         key  = 'SR_{}'.format(list(rating.keys())[0])
         val  = list(rating.values())[0]
         stats[key] = val.mu
     
      return stat_dict


###############################################################################
### Logging
class RLlibLogCallbacks(DefaultCallbacks):
   def on_episode_end(self, *, worker, base_env, policies, episode, **kwargs):
      assert len(base_env.envs) == 1, 'One env per worker'
      env    = base_env.envs[0]

      logs = env.terminal()
      for key, vals in logs['Stats'].items():
         episode.custom_metrics[key] = np.mean(vals)

      if not env.config.EVALUATE:
         return 

      agents = defaultdict(list)

      stats      = logs['Stats']
      policy_ids = stats['PolicyID']
      scores     = stats['Achievement']

      invMap = {agent.policyID: agent for agent in env.config.AGENTS}

      for policyID, score in zip(policy_ids, scores):
         policy = invMap[policyID]
         agents[policy].append(score)
      
      for agent in agents:
         agents[agent] = np.mean(agents[agent])

      policies = list(agents.keys())
      scores   = list(agents.values())

      idxs     = np.argsort(-np.array(scores))

      for rank, idx in enumerate(idxs):
          key = 'Rank_{}'.format(policies[idx].__name__)
          episode.custom_metrics[key] = rank




from pdb import set_trace as T
import numpy as np

from neural_mmo.forge.blade.lib import material
from neural_mmo.forge.blade.io.stimulus import Static

class Tile:
   def __init__(self, config, realm, r, c):
      self.config = config
      self.realm  = realm

      self.serialized = 'R{}-C{}'.format(r, c)

      self.r     = Static.Tile.R(realm.dataframe, self.serial, r)
      self.c     = Static.Tile.C(realm.dataframe, self.serial, c)
      self.nEnts = Static.Tile.NEnts(realm.dataframe, self.serial)
      self.index = Static.Tile.Index(realm.dataframe, self.serial, 0)

      realm.dataframe.init(Static.Tile, self.serial, (r, c))

   @property
   def serial(self):
      return self.serialized

   @property
   def repr(self):
      return ((self.r, self.c))

   @property
   def pos(self):
      return self.r.val, self.c.val

   @property
   def habitable(self):
      return self.mat in material.Habitable

   @property
   def vacant(self):
      return len(self.ents) == 0 and self.habitable

   @property
   def occupied(self):
      return not self.vacant

   @property
   def impassible(self):
      return self.mat in material.Impassible

   @property
   def lava(self):
      return self.mat == material.Lava

   @property
   def static(self):
      '''No updates needed'''
      assert self.capacity <= self.mat.capacity
      return self.capacity == self.mat.capacity

   def reset(self, mat, config):
      self.state  = mat(config)
      self.mat    = mat(config)

      self.capacity = self.mat.capacity
      self.tex      = mat.tex
      self.ents     = {}

      self.nEnts.update(0)
      self.index.update(self.state.index)
 
   def addEnt(self, ent):
      assert ent.entID not in self.ents
      self.nEnts.update(1)
      self.ents[ent.entID] = ent

   def delEnt(self, entID):
      assert entID in self.ents
      self.nEnts.update(0)
      del self.ents[entID]

   def step(self):
      if (not self.static and 
            np.random.rand() < self.mat.respawn):
         self.capacity += 1

      if self.static:
         self.state = self.mat
         self.index.update(self.state.index)

   def harvest(self):
      if self.capacity == 0:
         return False
      elif self.capacity <= 1:
         self.state = self.mat.degen(self.config)
         self.index.update(self.state.index)
      self.capacity -= 1
      return True
      return self.mat.dropTable.roll()



from pdb import set_trace as T
import ray, os
import time

OBJECT_INFO_PREFIX = b"OI:"
OBJECT_LOCATION_PREFIX = b"OL:"
TASK_TABLE_PREFIX = b"TT:"

def init(config, mode):
   os.environ['MKL_NUM_THREADS'] = '1'
   os.environ['OMP_NUM_THREADS'] = '1'

   if mode == 'local':
      ray.init(local_mode=True)
   elif mode == 'default':
      ray.init()
   elif mode == 'remote':
      ray.init(redis_address=config.HOST + ':6379')
      print('Cluster started with resources:')
      print(ray.cluster_resources())
   
   else:
      print('Invalid ray mode (local/default/remote)')
      exit(-1)

def clearbuffers():
   start = time.time()
   print('Clearing ray buffers...')
   flush_task_and_object_metadata_unsafe()
   flush_redis_unsafe()
   print('Done. Free time: ', time.time() - start)

#Continuous moving average
class CMA():
   def __init__(self):
      self.t = 1.0
      self.cma = None

   def update(self, x):
      if self.cma is None:
         self.cma = x
         return
      self.cma = (x + self.t*self.cma)/(self.t+1)
      self.t += 1.0

#Continuous moving average
class CMV():
   def __init__(self):
      self.cma = CMA()
      self.cmv = None

   def update(self, x):
      if self.cmv is None:
         self.cma.update(x)
         self.cmv = 0
         return
      prevMean = self.cma.cma
      self.cma.update(x)
      self.cmv += (x-prevMean)*(x-self.cma.cma)

   @property
   def stats(self):
      return self.cma.cma, self.cmv

class RayBenchmark:
   def __init__(self):
      self.cmv = CMV()
   def startRecord(self):
      self.start = time.time()

   def stopRecord(self):
      delta = time.time() - self.start
      self.cmv.update(delta)

   def reset(self):
      self.cmv = CMV()
      
   @property
   def stats(self):
      mean, var = self.cmv.stats
      return {'mean': mean, 'var': var}

def put(*args, profile=None):
   if profile is not None:
      if not hasattr(put, 'bench'):
         put.bench = {}
      if not profile in put.bench:
         put.bench[profile] = RayBenchmark()
      put.bench[profile].startRecord()
   ret = ray.put(*args)
   if profile is not None:
      put.bench[profile].stopRecord()
   return ret

def get(*args, profile=None):
   if profile is not None:
      if not hasattr(get, 'bench'):
         get.bench = {}
      if not profile in get.bench:
         get.bench[profile] = RayBenchmark()
      get.bench[profile].startRecord()
   ret = ray.get(*args)
   if profile is not None:
      get.bench[profile].stopRecord()
   return ret

def profile():
   stats = {}
   for k, f in zip('put get'.split(), (put, get)):
      if hasattr(f, 'bench'):
         stats[k] = dict((k, v.stats) for k, v in put.bench.items())
   return stats

def flush_redis_unsafe():
   """This removes some non-critical state from the primary Redis shard.

   This removes the log files as well as the event log from Redis. This can
   be used to try to address out-of-memory errors caused by the accumulation
   of metadata in Redis. However, it will only partially address the issue as
   much of the data is in the task table (and object table), which are not
   flushed.
   """
   if not hasattr(ray.worker.global_worker, "redis_client"):
       raise Exception("ray.experimental.flush_redis_unsafe cannot be called "
                       "before ray.init() has been called.")

   redis_client = ray.worker.global_worker.redis_client

   # Delete the log files from the primary Redis shard.
   keys = redis_client.keys("LOGFILE:*")
   if len(keys) > 0:
       num_deleted = redis_client.delete(*keys)
   else:
       num_deleted = 0
   print("Deleted {} log files from Redis.".format(num_deleted))

   # Delete the event log from the primary Redis shard.
   keys = redis_client.keys("event_log:*")
   if len(keys) > 0:
       num_deleted = redis_client.delete(*keys)
   else:
       num_deleted = 0
   print("Deleted {} event logs from Redis.".format(num_deleted))


def flush_task_and_object_metadata_unsafe():
   """This removes some critical state from the Redis shards.

   This removes all of the object and task metadata. This can be used to try
   to address out-of-memory errors caused by the accumulation of metadata in
   Redis. However, after running this command, fault tolerance will most
   likely not work.
   """
   if not hasattr(ray.worker.global_worker, "redis_client"):
       raise Exception("ray.experimental.flush_redis_unsafe cannot be called "
                       "before ray.init() has been called.")

   def flush_shard(redis_client):
       num_task_keys_deleted = 0
       for key in redis_client.scan_iter(match=TASK_TABLE_PREFIX + b"*"):
           num_task_keys_deleted += redis_client.delete(key)
       print("Deleted {} task keys from Redis.".format(num_task_keys_deleted))

       num_object_keys_deleted = 0
       for key in redis_client.scan_iter(match=OBJECT_INFO_PREFIX + b"*"):
           num_object_keys_deleted += redis_client.delete(key)
       print("Deleted {} object info keys from Redis.".format(
                 num_object_keys_deleted))

       num_object_location_keys_deleted = 0
       for key in redis_client.scan_iter(match=OBJECT_LOCATION_PREFIX + b"*"):
           num_object_location_keys_deleted += redis_client.delete(key)
       print("Deleted {} object location keys from Redis.".format(
                 num_object_location_keys_deleted))

   for redis_client in ray.worker.global_state.redis_clients:
       flush_shard(redis_client)



from neural_mmo.forge.ethyr.torch.policy.attention import Attention, SelfAttention
from neural_mmo.forge.ethyr.torch.policy.recurrent import BatchFirstLSTM
from neural_mmo.forge.ethyr.torch.policy.embed import BiasedInput, MixedDTypeInput
from neural_mmo.forge.ethyr.torch.policy.embed import Input



from pdb import set_trace as T

from enum import Enum, auto
from neural_mmo.forge.blade.lib.utils import staticproperty
from neural_mmo.forge.blade.io.comparable import IterableTypeCompare

class NodeType(Enum):
   #Tree edges
   STATIC = auto()    #Traverses all edges without decisions 
   SELECTION = auto() #Picks an edge to follow

   #Executable actions
   ACTION    = auto() #No arguments
   CONSTANT  = auto() #Constant argument
   VARIABLE  = auto() #Variable argument

class Node(metaclass=IterableTypeCompare):
   SERIAL = 2

   @staticproperty
   def edges():
      return []

   #Fill these in
   @staticproperty
   def priority():
      return None

   @staticproperty
   def type():
      return None

   @staticproperty
   def leaf():
      return False

   @classmethod
   def N(cls, config):
      return len(cls.edges)

   def args(stim, entity, config):
      return []



from pdb import set_trace as T
import numpy as np
import inspect
import os

from collections import defaultdict
from itertools import chain
import neural_mmo

class SequentialLoader:
    def __init__(self, items):
        for idx, itm in enumerate(items):
           itm.policyID = idx 

        self.items = items
        self.idx   = -1

    def __iter__(self):
        return self

    def __next__(self):
        self.idx = (self.idx + 1) % len(self.items)
        return self.items[self.idx]


class StaticIterable(type):
   def __iter__(cls):
      stack = list(cls.__dict__.items())
      for name, attr in cls.__dict__.items():
         if name == '__module__':
            continue
         if name.startswith('__'):
            break
         yield name, attr

class Template(metaclass=StaticIterable):
   def __init__(self):
      self.data = {}
      cls       = type(self)

      #Set defaults from static properties
      for k, v in cls:
         self.set(k, v)

   def override(self, **kwargs):
      for k, v in kwargs.items():
         err = 'CLI argument: {} is not a Config property'.format(k)
         assert hasattr(self, k), err
         self.set(k, v)

   def set(self, k, v):
      if type(v) is not property:
         try:
            setattr(self, k, v)
         except:
            print('Cannot set attribute: {} to {}'.format(k, v))
            quit()
      self.data[k] = v

   def print(self):
      keyLen = 0
      for k in self.data.keys():
         keyLen = max(keyLen, len(k))

      print('Configuration')
      for k, v in self.data.items():
         print('   {:{}s}: {}'.format(k, keyLen, v))

class Config(Template):
   '''An environment configuration object

   Global constants are defined as static class variables. You can override
   any Config variable using standard CLI syntax (e.g. --NENT=128). 

   The default config as of v1.5 uses 1024x1024 maps with up to 2048 agents
   and 1024 NPCs. It is suitable to time horizons of 8192+ steps. For smaller
   experiments, consider the SmallMaps config.
   
   Notes:
      We use Google Fire internally to replace standard manual argparse
      definitions for each Config property. This means you can subclass
      Config to add new static attributes -- CLI definitions will be
      generated automatically.
   '''
   ############################################################################
   ### Meta-Parameters
   ENV_NAME                = 'Neural_MMO'
   '''Environment Name'''

   ENV_VERSION             = '1.5.1'
   '''Environment version'''

   NAME_PREFIX             = 'Neural_'
   '''Prefix used in agent names displayed by the client'''

   v                       = False
   '''Verbose mode'''

   def game_system_enabled(self, name) -> bool:
      return hasattr(self, name)

   ############################################################################
   ### Population Parameters                                                   
   AGENTS                  = []

   AGENT_LOADER            = SequentialLoader

   NTILE                   = 6
   #TODO: Find a way to auto-compute this
   '''Number of distinct terrain tile types'''

   NSTIM                   = 7
   '''Number of tiles an agent can see in any direction'''

   NMOB                    = 1024
   '''Maximum number of NPCs spawnable in the environment'''

   NENT                    = 2048
   '''Maximum number of agents spawnable in the environment'''

   NPOP                    = 1
   '''Number of distinct populations spawnable in the environment'''

   COOPERATIVE             = False
   '''Whether to treat populations as teams'''

   @property
   def TEAM_SIZE(self):
      assert not self.NMOB % self.NPOP
      return self.NMOB // self.NPOP

   @property
   def WINDOW(self):
      '''Size of the square tile crop visible to an agent'''
      return 2*self.NSTIM + 1

   REWARD_ACHIEVEMENT      = False

   ############################################################################
   ### Agent Parameters                                                   
   BASE_HEALTH                = 10
   '''Initial Constitution level and agent health'''

   BASE_RESOURCE              = 20
   '''Initial level and capacity for Hunting + Fishing resource skills'''


   PLAYER_SPAWN_ATTEMPTS   = 16
   '''Number of player spawn attempts per tick

   Note that the env will attempt to spawn agents until success
   if the current population size is zero.'''

   def SPAWN_CONTINUOUS(self):
      '''Generates spawn positions for new agents

      Default behavior randomly selects a tile position
      along the borders of the square game map

      Returns:
         tuple(int, int):

         position:
            The position (row, col) to spawn the given agent
      '''
      #Spawn at edges
      mmax = self.TERRAIN_CENTER + self.TERRAIN_BORDER
      mmin = self.TERRAIN_BORDER

      var  = np.random.randint(mmin, mmax)
      fixed = np.random.choice([mmin, mmax])
      r, c = int(var), int(fixed)
      if np.random.rand() > 0.5:
          r, c = c, r 
      return (r, c)

   def SPAWN_CONCURRENT(self):
      left   = self.TERRAIN_BORDER
      right  = self.TERRAIN_CENTER + self.TERRAIN_BORDER
      rrange = np.arange(left+2, right, 4).tolist()

      assert not self.TERRAIN_CENTER % 4
      per_side = self.TERRAIN_CENTER // 4
      
      lows   = (left+np.zeros(per_side, dtype=np.int)).tolist()
      highs  = (right+np.zeros(per_side, dtype=np.int)).tolist()

      s1     = list(zip(rrange, lows))
      s2     = list(zip(lows, rrange))
      s3     = list(zip(rrange, highs))
      s4     = list(zip(highs, rrange))

      return s1 + s2 + s3 + s4

   @property
   def SPAWN(self):
      return self.SPAWN_CONTINUOUS

   ############################################################################
   ### Evaluation Parameters
   EVALUATE             = False
   '''Flag used by evaluation mode'''

   GENERALIZE           = True
   '''Evaluate on maps not seen during training'''

   RENDER               = False
   '''Flag used by render mode'''

   ############################################################################
   ### Terrain Generation Parameters
   TERRAIN_TRAIN_MAPS            = 256
   '''Number of training maps to generate'''

   TERRAIN_EVAL_MAPS             = 64
   '''Number of evaluation maps to generate'''

   TERRAIN_RENDER             = False
   '''Whether map generation should also save .png previews (slow + large file size)'''

   TERRAIN_CENTER             = 1024
   '''Size of each map (number of tiles along each side)'''

   TERRAIN_BORDER             = 16
   '''Number of lava border tiles surrounding each side of the map'''

   @property
   def TERRAIN_SIZE(self):
      return int(self.TERRAIN_CENTER + 2*self.TERRAIN_BORDER)

   TERRAIN_FREQUENCY          = -3
   '''Base noise frequency range (log2 space)'''

   TERRAIN_FREQUENCY_OFFSET   = 7
   '''Noise frequency octave offset (log2 space)'''

   TERRAIN_LOG_INTERPOLATE_MIN = -2 
   '''Minimum interpolation log-strength for noise frequencies'''

   TERRAIN_LOG_INTERPOLATE_MAX= 0
   '''Maximum interpolation log-strength for noise frequencies'''

   TERRAIN_TILES_PER_OCTAVE   = 8
   '''Number of octaves sampled from log2 spaced TERRAIN_FREQUENCY range'''

   TERRAIN_LAVA               = 0.0
   '''Noise threshold for lava generation'''

   TERRAIN_WATER              = 0.30
   '''Noise threshold for water generation'''

   TERRAIN_GRASS              = 0.70
   '''Noise threshold for grass'''

   TERRAIN_FOREST             = 0.85
   '''Noise threshold for forest'''


   ############################################################################
   ### Visualization Parameters
   VIS_THEME            = 'web'
   '''Visualizer theme: web or publication'''

   VIS_WIDTH            = 1920
   '''Visualizer figure width (pixels)'''

   VIS_HEIGHT           = 314
   '''Visualizer per-plot height (pixels)'''

   VIS_BORDER_WIDTH     = 20
   '''Horizontal padding per figure side (pixels)'''

   VIS_BORDER_HEIGHT    = 60
   '''Vertical padding per figure side (pixels)'''

   VIS_LEGEND_WIDTH     = 109
   '''Width of legend label before offset'''
   
   VIS_LEGEND_OFFSET    = 71 
   '''Width of legend label offset'''

   VIS_TITLE_OFFSET     = 60
   '''Width of left title offset'''

   VIS_PORT             = 5006
   '''Visualizer local Bokeh server port'''

   VIS_TOOLS            = False
   '''Visualizer display plot tools'''


   ############################################################################
   ### Path Parameters
   #PATH_ROOT            = os.path.dirname(neural_mmo.__file__)
   PATH_ROOT            = os.getcwd()
   '''Global repository directory'''

   PATH_RESOURCE        = os.path.join(PATH_ROOT, 'resource')
   '''Resource directory'''

   PATH_MAPS            = os.path.join(PATH_RESOURCE, 'maps')
   '''Generated map directory'''

   PATH_MAP_SUFFIX      = 'map{}/map.npy'
   '''Map file name'''

   PATH_MAPS_SMALL      = os.path.join(PATH_MAPS, 'procedural-small')
   '''Generated map directory for SmallMap config'''

   PATH_MAPS_LARGE      = os.path.join(PATH_MAPS, 'procedural-large')
   '''Generated map directory for LargeMap config'''

   PATH_MAPS            = PATH_MAPS_LARGE
   '''Generated map directory'''


   #Assets
   PATH_ASSETS          = os.path.join(PATH_RESOURCE, 'assets')
   '''Asset directory'''

   PATH_TILE            = os.path.join(PATH_ASSETS, 'tiles/{}.png')
   '''Tile path -- format me with tile name'''

   PATH_TEXT            = os.path.join(PATH_ASSETS, 'text')
   '''Text directory'''

   PATH_LOGO            = os.path.join(PATH_TEXT, 'ascii.txt')
   '''Logo file (Ascii art)'''


   #Baselines and Checkpoints
   PATH_CHECKPOINTS          = 'checkpoints'
   '''Checkpoints path'''

   PATH_BASELINES            = 'baselines'
   '''Model and evaluation directory'''
  
   PATH_ALL_MODELS           = os.path.join(PATH_BASELINES, 'models')
   '''All models directory'''

   @property
   def PATH_MODEL(self):
      '''Model path'''
      return os.path.join(self.PATH_ALL_MODELS, self.MODEL)

   @property
   def PATH_TRAINING_DATA(self):
      '''Model training data'''
      return os.path.join(self.PATH_MODEL, 'training.npy')

   PATH_ALL_EVALUATIONS      = os.path.join(PATH_BASELINES, 'evaluations')
   '''All evaluations directory'''

   @property
   def PATH_EVALUATION(self):
      '''Evaluation path'''
      return os.path.join(self.PATH_ALL_EVALUATIONS, self.MODEL + '.npy')

   #Themes
   PATH_THEMES          = os.path.join('neural_mmo', 'forge', 'blade', 'systems', 'visualizer') 
   '''Theme directory'''

   PATH_THEME_WEB       = os.path.join(PATH_THEMES, 'index_web.html')
   '''Web theme file'''

   PATH_THEME_PUB       = os.path.join(PATH_THEMES, 'index_publication.html')
   '''Publication theme file'''

class SmallMaps(Config):
   '''A smaller config modeled off of v1.4 and below featuring 128x128 maps
   with up to 256 agents and 128 NPCs. It is suitable to time horizons of 1024+
   steps. For larger experiments, consider the default config.'''

   PATH_MAPS               = Config.PATH_MAPS_SMALL

   #Scale
   TERRAIN_CENTER          = 128
   NENT                    = 256
   NMOB                    = 128

   #Players spawned per tick
   PLAYER_SPAWN_ATTEMPTS   = 2

   #NPC parameters
   NPC_LEVEL_MAX           = 30
   NPC_LEVEL_SPREAD        = 5


############################################################################
### Game Systems (Static Mixins)
class Resource:
   '''Resource Game System'''

   @property #Reserved flag
   def Resource(self):
      return True

   RESOURCE_BASE_RESOURCE              = 25
   '''Initial level and capacity for Hunting + Fishing resource skills'''

   RESOURCE_FOREST_CAPACITY            = 1
   '''Maximum number of harvests before a forest tile decays'''

   RESOURCE_FOREST_RESPAWN             = 0.025
   '''Probability that a harvested forest tile will regenerate each tick'''

   RESOURCE_OREROCK_CAPACITY           = 1
   '''Maximum number of harvests before an orerock tile decays'''

   RESOURCE_OREROCK_RESPAWN            = 0.05
   '''Probability that a harvested orerock tile will regenerate each tick'''

   RESOURCE_HARVEST_RESTORE_FRACTION   = 1.0
   '''Fraction of maximum capacity restored upon collecting a resource'''

   RESOURCE_HEALTH_REGEN_THRESHOLD     = 0.5
   '''Fraction of maximum resource capacity required to regen health'''

   RESOURCE_HEALTH_RESTORE_FRACTION    = 0.1
   '''Fraction of health restored per tick when above half food+water'''

class Combat:
   '''Combat Game System'''

   @property #Reserved flag
   def Combat(self):
      return True

   COMBAT_DICE_SIDES                   = 20
   '''Number of sides for combat dice

   Attacks can only hit opponents up to the attacker's level plus
   DICE_SIDES/2. Increasing this value makes attacks more accurate
   and allows lower level attackers to hit stronger opponents'''

   COMBAT_DEFENSE_WEIGHT               = 0.3 
   '''Fraction of defense that comes from the Defense skill'''   

   COMBAT_MELEE_REACH                  = 1
   '''Reach of attacks using the Melee skill'''

   COMBAT_RANGE_REACH                  = 3
   '''Reach of attacks using the Range skill'''

   COMBAT_MAGE_REACH                   = 4
   '''Reach of attacks using the Mage skill'''

   COMBAT_FREEZE_TIME                  = 3
   '''Number of ticks successful Mage attacks freeze a target'''


class Progression:
   '''Progression Game System'''

   @property #Reserved flag
   def Progression(self):
      return True

   PROGRESSION_BASE_RESOURCE           = 10
   '''Initial level and capacity for Hunting + Fishing resource skills'''

   PROGRESSION_BASE_XP_SCALE           = 10
   '''Skill level progression speed as a multiplier of typical MMOs'''

   PROGRESSION_CONSTITUTION_XP_SCALE   = 2
   '''Multiplier on top of XP_SCALE for the Constitution skill'''

   PROGRESSION_COMBAT_XP_SCALE         = 4
   '''Multiplier on top of XP_SCALE for Combat skills'''


class NPC:
   '''NPC & Equipment Game System'''

   @property #Reserved flag
   def NPC(self):
      return True

   NPC_SPAWN_ATTEMPTS                  = 25
   '''Number of NPC spawn attempts per tick'''

   NPC_SPAWN_AGGRESSIVE                = 0.80
   '''Percentage distance threshold from spawn for aggressive NPCs'''

   NPC_SPAWN_NEUTRAL                   = 0.50
   '''Percentage distance threshold from spawn for neutral NPCs'''

   NPC_SPAWN_PASSIVE                   = 0.00
   '''Percentage distance threshold from spawn for passive NPCs'''
   
   NPC_LEVEL_MIN                       = 1
   '''Minimum NPC level'''

   NPC_LEVEL_MAX                       = 99
   '''Maximum NPC level'''

   NPC_LEVEL_SPREAD                    = 10 
   '''Level range for NPC spawns'''

class Achievement:
   '''Achievement Reward System'''

   @property #Reserved flag
   def Achievement(self):
      return True

   PLAYER_KILLS_EASY       = 1
   PLAYER_KILLS_NORMAL     = 3
   PLAYER_KILLS_HARD       = 6

   EQUIPMENT_EASY          = 1
   EQUIPMENT_NORMAL        = 10
   EQUIPMENT_HARD          = 20

   EXPLORATION_EASY        = 32
   EXPLORATION_NORMAL      = 64
   EXPLORATION_HARD        = 127

   FORAGING_EASY           = 20
   FORAGING_NORMAL         = 35
   FORAGING_HARD           = 50

class AllGameSystems(Resource, Combat, Progression, NPC): pass



from pdb import set_trace as T

from neural_mmo.forge.blade.lib import enums

class Agent:
    scripted = False
    name     = 'Neural_'
    color    = enums.Neon.CYAN
    pop      = 0

    def __init__(self, config, idx):
       self.config = config
       self.iden   = idx
       self.pop    = Agent.pop





from pdb import set_trace as T
from collections import defaultdict
from neural_mmo.forge.blade.lib import material
from copy import deepcopy
import os

import numpy as np
import json, pickle
import time
import ray

from neural_mmo.forge.blade.lib import utils
from neural_mmo.forge.blade.systems import visualizer
from neural_mmo.forge.blade.systems.visualizer import plot

class Quill:
   TRAINING     = -1
   LINE         = 0
   SCATTER      = 1
   HISTOGRAM    = 2
   GANTT        = 3
   STATS        = 4
   RADAR        = 5
   STACKED_AREA = 6

   PLOTS = {
      TRAINING:     plot.Training,
      LINE:         plot.Line,
      SCATTER:      plot.Scatter,
      HISTOGRAM:    plot.Histogram,
      GANTT:        plot.Gantt,
      STATS:        plot.Stats,
      RADAR:        plot.Radar,
      STACKED_AREA: plot.StackedArea,
   }

   def plot(idx):
      return Quill.PLOTS[idx] 

   def __init__(self):
      self.blobs = defaultdict(Blob)
      self.stats = defaultdict(list)

   def stat(self, key, val):
      self.stats[key].append(val)

   def register(self, key, tick, *plots):
      if key in self.blobs:
         blob = self.blobs[key]
      else:
         blob = Blob(*plots)
         self.blobs[key] = blob

      blob.tick = tick
      return blob

   @property
   def packet(self):
      logs = {key: blob.packet for key, blob in self.blobs.items()}
      return {'Log': logs, 'Stats': self.stats}

class Blob:
   def __init__(self, *args):
      self.tracks = defaultdict(Track)
      self.plots  = args
 
   def log(self, value, key=None):
      self.tracks[key].update(self.tick, value)

   @property
   def packet(self):
      return {key: (self.plots, track.packet) for key, track in self.tracks.items()}

#Static blob analytics
class Track:
   def __init__(self):
      self.data = defaultdict(list)

   def update(self, tick, value):
      if type(value) != list:
         value = [value]

      self.data[tick] += value

   @property
   def packet(self):
      return self.data

class InkWell:
   def __init__(self):
      self.log   = defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(list))))
      self.stats = defaultdict(list)

   def update(self, quill, realm=0):
      '''Realm param unused -- for future interactive expansions'''
      log, stats = quill['Log'], quill['Stats']
      for key, blob in log.items():
         for subkey, (plots, track) in blob.items():
            for time, vals in track.items():
               self.log[realm][key, tuple(plots)][subkey][time] += vals
      for key, stat in stats.items():
         self.stats[key] += stat

   @property
   def packet(self):
      return {'Log': utils.default_to_regular(self.log),
              'Stats': utils.default_to_regular(self.stats)}
 
#Log wrapper and benchmarker
class Benchmarker:
   def __init__(self, logdir):
      self.benchmarks = {}

   def wrap(self, func):
      self.benchmarks[func] = Utils.BenchmarkTimer()
      def wrapped(*args):
         self.benchmarks[func].startRecord()
         ret = func(*args)
         self.benchmarks[func].stopRecord()
         return ret
      return wrapped

   def bench(self, tick):
      if tick % 100 == 0:
         for k, benchmark in self.benchmarks.items():
            bench = benchmark.benchmark()
            print(k.__func__.__name__, 'Tick: ', tick,
                  ', Benchmark: ', bench, ', FPS: ', 1/bench)
 




from pdb import set_trace as T
import numpy as np

import torch
from torch import nn

class ScaledDotProductAttention(nn.Module):
   def __init__(self, h):
      super().__init__()
      self.scale = np.sqrt(h)

   def forward(self, Q, K, V):
      Kt = K.transpose(-2, -1)
      QK = torch.matmul(Q, Kt)
      QK = QK / self.scale
      #QK = torch.softmax(QK / self.scale, -1)
      #QK = torch.nn.functional.softmax(QK, dim=-2)
      QKV = torch.matmul(QK, V)
      return QKV

class MultiHeadAttention(nn.Module):
   def __init__(self, h, nHeads):
      super().__init__()
      self.headDim = int(h / nHeads)
      self.nHeads = nHeads

      self.Q = nn.Linear(h, h)
      self.K = nn.Linear(h, h)
      self.V = nn.Linear(h, h)

      self.attention = ScaledDotProductAttention(h)

   def reperm(self, x, src, perm, dst):
      return  x.view(*src).permute(*perm).reshape(*dst)

   def forward(self, q, kv=None):
      if kv is None:
         k = v = q
      else:
         k = v = kv

      assert k.shape == v.shape

      perm = (3, 0, 1, 2, 4)
      headDim, nHead = self.headDim, self.nHeads

      batch1, batch2, seq, _ = k.size()
      src  = (batch1, batch2, seq, nHead, headDim)
      dst  = (batch1*batch2*nHead, seq, headDim)

      k = self.reperm(self.K(k), src, perm, dst)
      v = self.reperm(self.V(v), src, perm, dst)

      batch1, batch2, seq, _ = q.size()
      src  = (batch1, batch2, seq, nHead, headDim)
      dst  = (batch1*batch2*nHead, seq, headDim)

      q = self.reperm(self.Q(q), src, perm, dst)

      perm = (1, 2, 0, 3)
      src  = (nHead, batch1*batch2, seq, headDim)
      dst  = (batch1, batch2, seq, headDim*nHead)

      attn = self.attention(q, k, v)
      attn = self.reperm(attn, src, perm, dst)

      return attn

class Block(nn.Module):
   def __init__(self, h, nHeads, norm=False):
      super().__init__()
      self.attn = MultiHeadAttention(h, nHeads)
      self.fc   = nn.Linear(h, h)

      self.normalize = norm
      if norm:
         self.norm = nn.LayerNorm(h)

   def forward(self, x, kv=None):
      x = self.attn(x, kv) + x
      if self.normalize:
         x = self.norm(x)

      x = self.fc(x) + x
      if self.normalize:
         x = self.norm(x)

      return x

class Transformer(nn.Module):
   def __init__(self, h, nHeads, nLayers=1, flat=True):
      super().__init__()
      modules = [Block(h, nHeads) for i in range(nLayers)]
      self.attns = nn.ModuleList(modules)
      self.flat = flat

   def forward(self, x, kv=None):
      for attn in self.attns:
         x = attn(x, kv)

      if self.flat:
         x = x.mean(-2)

      return x



from pdb import set_trace as T
import numpy as np
import gym

import inspect
from enum import Enum, auto
from collections import deque

from neural_mmo.forge.blade.lib.utils import classproperty, staticproperty

class Iterable(type):
   def __iter__(cls):
      queue = deque(cls.__dict__.items())
      while len(queue) > 0:
         name, attr = queue.popleft()
         if type(name) != tuple:
            name = tuple([name])
         if not inspect.isclass(attr):
            continue
         if issubclass(attr, Flat):
            for n, a in attr.__dict__.items():
               n = name + tuple([n])
               queue.append((n, a))
            continue
         yield name, attr

   def values(cls):
      return [e[1] for e in cls]

class NameComparable(type):
   def __hash__(self):
      return hash(self.__name__)

   def __eq__(self, other):
      return self.__name__ == other.__name__

   def __ne__(self, other):
      return self.__name__ != other.__name__

   def __lt__(self, other):
      return self.__name__ < other.__name__

   def __le__(self, other):
      return self.__name__ <= other.__name__

   def __gt__(self, other):
      return self.__name__ > other.__name__

   def __ge__(self, other):
      return self.__name__ >= other.__name__

class IterableNameComparable(Iterable, NameComparable):
   pass

class Flat:
   pass

class Stim:
   CONTINUOUS = False
   DISCRETE   = False
   def __init__(self, dataframe, key, val=None, config=None):
      if config is None:
         config    = dataframe.config
 
      self.obj  = str(self.__class__).split('.')[-2]
      self.attr = self.__class__.__name__
      self.key  = key

      self.min = 0
      self.max = np.inf
      self.val = val

      self.dataframe = dataframe
      self.init(config)
      err = 'Must set a default val upon instantiation or init()'
      assert self.val is not None, err

      #Update dataframe
      if dataframe is not None:
         self.update(self.val)

   #Defined for cleaner stim files
   def init(self):
      pass

   def packet(self):
      return {
            'val': self.val,
            'max': self.max}

   def update(self, val):
      self.val = min(max(val, self.min), self.max)
      self.dataframe.update(self, self.val)
      return self

   def increment(self, val=1):
      self.update(self.val + val)
      return self

   def decrement(self, val=1):
      self.update(self.val - val)
      return self

   @property
   def empty(self):
      return self.val == 0

   def __add__(self, other):
      self.increment(other)
      return self

   def __sub__(self, other):
      self.decrement(other)
      return self

   def __eq__(self, other):
      return self.val == other

   def __ne__(self, other):
      return self.val != other

   def __lt__(self, other):
      return self.val < other

   def __le__(self, other):
      return self.val <= other

   def __gt__(self, other):
      return self.val > other

   def __ge__(self, other):
      return self.val >= other

class Continuous(Stim):
   CONTINUOUS = True

class Discrete(Continuous):
   DISCRETE = True

class NodeType(Enum):
   #Tree edges
   STATIC = auto()    #Traverses all edges without decisions
   SELECTION = auto() #Picks an edge to follow

   #Executable actions
   ACTION    = auto() #No arguments
   CONSTANT  = auto() #Constant argument
   VARIABLE  = auto() #Variable argument

class Node(metaclass=IterableNameComparable):
   @staticproperty
   def edges():
      return []

   #Fill these in
   @staticproperty
   def priority():
      return None

   @staticproperty
   def type():
      return None

   @staticproperty
   def leaf():
      return False

   @classmethod
   def N(cls, config):
      return len(cls.edges)

   def args(stim, entity, config):
      return []



from pdb import set_trace as T
import numpy as np

import torch
from torch import nn
from torch.nn import functional as F

def Conv2d(fIn, fOut, k, stride=1):
   '''torch Conv2d with same padding'''
   assert k % 2 == 0
   pad = int((k-1)/2)
   return torch.nn.Conv2d(fIn, fOut, 
      k, stride=stride, padding=pad)

def Pool(k, stride=1, pad=0):
   return torch.nn.MaxPool2d(
      k, stride=stride, padding=pad)

class ConvReluPool(nn.Module):
   def __init__(self, fIn, fOut, 
         k, stride=1, pool=2):
      super().__init__()
      self.conv = Conv2d(
         fIn, fOut, k, stride)
      self.pool = Pool(k)

   def forward(self, x):
      x = self.conv(x)
      x = F.relu(x)
      x = self.pool(x)
      return x




#Various utilities for managing combat, including hit/damage

from pdb import set_trace as T

import numpy as np
from neural_mmo.forge.blade.systems import skill as Skill

def level(skills):
   hp = skills.constitution.level
   defense = skills.defense.level
   melee = skills.melee.level
   ranged = skills.range.level
   mage   = skills.mage.level
   
   base = 0.25*(defense + hp)
   final = np.floor(base + 0.5*max(melee, ranged, mage))
   return final

def attack(entity, targ, skillFn):
   config      = entity.config
   entitySkill = skillFn(entity)
   targetSkill = skillFn(targ)

   targetDefense = targ.skills.defense.level + targ.loadout.defense

   die  = config.COMBAT_DICE_SIDES
   roll = np.random.randint(1, die+1)
   dc   = accuracy(config, entitySkill.level, targetSkill.level, targetDefense)
   crit = roll == die

   dmg = 1 #Chip dmg on a miss
   if roll >= dc or crit:
      dmg = damage(entitySkill.__class__, entitySkill.level)
      
   dmg = min(dmg, entity.resources.health.val)
   entity.applyDamage(dmg, entitySkill.__class__.__name__.lower())
   targ.receiveDamage(entity, dmg)
   return dmg

#Compute maximum damage roll
def damage(skill, level):
   if skill == Skill.Melee:
      return np.floor(7 + level * 63 / 99)
   if skill == Skill.Range:
      return np.floor(3 + level * 32 / 99)
   if skill == Skill.Mage:
      return np.floor(1 + level * 24 / 99)

#Compute maximum attack or defense roll (same formula)
#Max attack 198 - min def 1 = 197. Max 198 - max 198 = 0
#REMOVE FACTOR OF 2 FROM ATTACK AFTER IMPLEMENTING WEAPONS
def accuracy(config, entAtk, targAtk, targDef):
   alpha   = config.COMBAT_DEFENSE_WEIGHT

   attack  = entAtk
   defense = alpha*targDef + (1-alpha)*targAtk
   dc      = defense - attack + config.COMBAT_DICE_SIDES//2

   return dc

def danger(config, pos, full=False):
   border = config.TERRAIN_BORDER
   center = config.TERRAIN_CENTER
   r, c   = pos
  
   #Distance from border
   rDist  = min(r - border, center + border - r - 1)
   cDist  = min(c - border, center + border - c - 1)
   dist   = min(rDist, cDist)
   norm   = 2 * dist / center

   if full:
      return norm, mag

   return norm



from pdb import set_trace as T
import numpy as np

from neural_mmo.forge.blade.systems import skill, droptable, combat, equipment
from neural_mmo.forge.blade.lib import material, utils

from neural_mmo.forge.blade.io.action import static as Action
from neural_mmo.forge.blade.io.stimulus import Static

class Resources:
   def __init__(self, ent):
      self.health = Static.Entity.Health(ent.dataframe, ent.entID)
      self.water  = Static.Entity.Water( ent.dataframe, ent.entID)
      self.food   = Static.Entity.Food(  ent.dataframe, ent.entID)

   def update(self, realm, entity, actions):
      self.health.max = entity.skills.constitution.level
      self.water.max  = entity.skills.fishing.level
      self.food.max   = entity.skills.hunting.level

   def packet(self):
      data = {}
      data['health'] = self.health.packet()
      data['food']   = self.food.packet()
      data['water']  = self.water.packet()
      return data

class Status:
   def __init__(self, ent):
      self.config = ent.config
      self.freeze     = Static.Entity.Freeze(ent.dataframe, ent.entID)

   def update(self, realm, entity, actions):
      self.freeze.decrement()

   def packet(self):
      data = {}
      data['freeze']     = self.freeze.val
      return data

class History:
   def __init__(self, ent):
      self.actions = None
      self.attack  = None
  
      self.origPos     = ent.pos
      self.exploration = 0
      self.playerKills = 0

      self.damage    = Static.Entity.Damage(   ent.dataframe, ent.entID)
      self.timeAlive = Static.Entity.TimeAlive(ent.dataframe, ent.entID)

      self.lastPos = None

   def update(self, realm, entity, actions):
      self.attack  = None
      self.actions = actions
      self.damage.update(0)

      exploration      = utils.linf(entity.pos, self.origPos)
      self.exploration = max(exploration, self.exploration)

      self.timeAlive.increment()

   def packet(self):
      data = {}
      data['damage']    = self.damage.val
      data['timeAlive'] = self.timeAlive.val

      if self.attack is not None:
         data['attack'] = self.attack

      return data

class Base:
   def __init__(self, ent, pos, iden, name, color, pop):
      self.name  = name + str(iden)
      self.color = color
      r, c       = pos

      self.r          = Static.Entity.R(ent.dataframe, ent.entID, r)
      self.c          = Static.Entity.C(ent.dataframe, ent.entID, c)

      self.population = Static.Entity.Population(ent.dataframe, ent.entID, pop)
      self.self       = Static.Entity.Self(      ent.dataframe, ent.entID, 1)
      self.identity   = Static.Entity.ID(        ent.dataframe, ent.entID, ent.entID)
      self.level      = Static.Entity.Level(     ent.dataframe, ent.entID, 3)

      ent.dataframe.init(Static.Entity, ent.entID, (r, c))

   def update(self, realm, entity, actions):
      self.level.update(combat.level(entity.skills))

   @property
   def pos(self):
      return self.r.val, self.c.val

   def packet(self):
      data = {}

      data['r']          = self.r.val
      data['c']          = self.c.val
      data['name']       = self.name
      data['color']      = self.color.packet()
      data['population'] = self.population.val
      data['self']       = self.self.val

      return data

class Entity:
   def __init__(self, realm, pos, iden, name, color, pop):
      self.dataframe    = realm.dataframe
      self.config       = realm.config
      self.entID        = iden

      self.repr         = None
      self.vision       = 5

      self.attacker     = None
      self.target       = None
      self.closest      = None
      self.spawnPos     = pos

      self.attackerID = Static.Entity.AttackerID(self.dataframe, self.entID, 0)

      #Submodules
      self.base      = Base(self, pos, iden, name, color, pop)
      self.status    = Status(self)
      self.history   = History(self)
      self.resources = Resources(self)
      self.loadout   = equipment.Loadout()

   def packet(self):
      data = {}

      data['status']  = self.status.packet()
      data['history'] = self.history.packet()
      data['loadout'] = self.loadout.packet()
      data['alive']   = self.alive

      return data

   def update(self, realm, actions):
      '''Update occurs after actions, e.g. does not include history'''
      if self.history.damage == 0:
         self.attacker = None
         self.attackerID.update(0)

      self.base.update(realm, self, actions)
      self.status.update(realm, self, actions)
      self.history.update(realm, self, actions)

   def receiveDamage(self, source, dmg):
      self.history.damage.update(dmg)
      self.resources.health.decrement(dmg)

      if not self.alive and source:
         source.receiveLoot(self.loadout)
         return False

      return True

   def receiveLoot(self, loadout):
      pass

   def applyDamage(self, dmg, style):
      pass

   @property
   def pos(self):
      return self.base.pos

   @property
   def alive(self):
      if self.resources.health.empty:
         return False

      return True

   @property
   def isPlayer(self) -> bool:
      return False

   @property
   def isNPC(self) -> bool:
      return False



from collections import defaultdict
from queue import PriorityQueue

class Offer:
   def __init__(self, item, quant, itemPrice, coffer=0):
      self.item = item
      self.quantLeft = quant
      self.quantFulfilled = 0
      self.itemPrice = itemPrice
      self.coffer = coffer

   @property
   def complete(self):
      return self.quantLeft == 0

   def partialCollect(self):
      pass

   def cancel(self):
      return self.coffer, [self.item() for e in self.quantFulfilled]

   def __lt__(self, other):
      return True

   def __eq__(self, other):
      return False

class BuyOffer(Offer):
   def buy(self, quant, itemPrice):
      self.coffer -= itemPrice * quant

   #Collect only profits thus far
   def partialCollect(self):
      ret = [self.item() for e in self.quantFulfilled]
      self.quantFulfilled = 0
      return ret

class SellOffer(Offer):
   def sell(self, quant):
      self.coffer += self.itemPrice * quant
      self.quantLeft -= quant
      assert self.quantLeft >= 0

   def partialCollect(self):
      ret = self.coffer
      self.coffer = 0
      return self.coffer

#Why is there no peek fuction...
class PQ(PriorityQueue):
   def peek(self):
      if len(self.queue) > 0:
         return self.queue[0]
      return None

class Exchange:
   def __init__(self):
      self.buyOffers  = defaultdict(PQ)
      self.sellOffers = defaultdict(PQ)

   def buy(self, item, quant, maxPrice):
      offer = BuyOffer(item, quant, maxPrice, coffer=quant*maxPrice)
      self.buyOffers[item].put(offer, -maxPrice)
      self.update(item)
      return offer

   def sell(self, item, quant, itemPrice):
      offer = SellOffer(item, quant, itemPrice)
      self.sellOffers[item].put(offer, itemPrice)
      self.update(item)
      return offer

   def update(self, item):
      buyOffer  = self.buyOffers[item].peek()
      sellOffer = self.sellOffers[item].peek()

      if None in (buyOffer, sellOffer):
         return
      maxBuy, minSell = buyOffer.itemPrice, sellOffer.itemPrice
      itemPrice = minSell #Advantage given to buyer arbitrarily

      if maxBuy >= minSell:
         if sellOffer.quantLeft < buyOffer.quantLeft:
            buyOffer.buy(sellOffer.quantLeft, itemPrice)
            sellOffer.sell(sellOffer.quantLeft)
            self.sellOffers[item].get()
         elif sellOffer.quantLeft > buyOffer.quantLeft:
            buyOffer.buy(buyOffer.quantLeft, itemPrice)
            sellOffer.sell(buyOffer.quantLeft)
            self.buyOffers[item].get()
         elif sellOffer.quantLeft == buyOffer.quantLeft:
            buyOffers.buy(buyOffer.quantLeft, itemPrice)
            sellOffer.sell(sellOffer.quantLeft)
            self.buyOffers[item].get()
            self.sellOffers[item].get()

         self.update(item)



from pdb import set_trace as T

from neural_mmo.forge.blade import core
from neural_mmo.forge.blade.core import config
from neural_mmo.forge.blade.io.stimulus.static import Stimulus
from neural_mmo.forge.trinity.scripted import baselines
from neural_mmo.forge.trinity.agent import Agent
from neural_mmo.forge.blade.systems.ai import behavior
from projekt import rllib_wrapper

class RLlibConfig:
   '''Base config for RLlib Models

   Extends core Config, which contains environment, evaluation,
   and non-RLlib-specific learning parameters'''

   @property
   def MODEL(self):
      return self.__class__.__name__
  
   #Policy specification
   AGENTS      = [Agent]
   EVAL_AGENTS = [baselines.Meander, baselines.Forage, baselines.Combat, Agent]
   EVALUATE    = False #Reserved param
   LOAD        = True

   #Hardware and debug
   NUM_WORKERS             = 1
   NUM_GPUS_PER_WORKER     = 0
   NUM_GPUS                = 1
   EVALUATION_NUM_WORKERS  = 3
   LOCAL_MODE              = False
   LOG_LEVEL               = 1

   #Training and evaluation settings
   EVALUATION_INTERVAL     = 1
   EVALUATION_NUM_EPISODES = 3
   EVALUATION_PARALLEL     = True
   TRAINING_ITERATIONS     = 1000
   KEEP_CHECKPOINTS_NUM    = 5
   CHECKPOINT_FREQ         = 1
   LSTM_BPTT_HORIZON       = 16
   NUM_SGD_ITER            = 1

   #Model
   SCRIPTED                = None
   N_AGENT_OBS             = 100
   NPOLICIES               = 1
   HIDDEN                  = 64
   EMBED                   = 64

   #Reward
   TEAM_SPIRIT             = 0.0
   ACHIEVEMENT_SCALE       = 1.0/15.0


class LargeMaps(core.Config, RLlibConfig, config.AllGameSystems):
   '''Large scale Neural MMO training setting

   Features up to 1000 concurrent agents and 1000 concurrent NPCs,
   1km x 1km maps, and 5/10k timestep train/eval horizons

   This is the default setting as of v1.5 and allows for large
   scale multiagent research even on relatively modest hardware'''

   #Memory/Batch Scale
   NUM_WORKERS             = 14
   TRAIN_BATCH_SIZE        = 32 * NUM_WORKERS
   ROLLOUT_FRAGMENT_LENGTH = 32
   SGD_MINIBATCH_SIZE      = 128

   #Horizon
   TRAIN_HORIZON           = 8192
   EVALUATION_HORIZON      = 8192


class SmallMaps(RLlibConfig, config.AllGameSystems, config.SmallMaps):
   '''Small scale Neural MMO training setting

   Features up to 128 concurrent agents and 32 concurrent NPCs,
   60x60 maps (excluding the border), and 1000 timestep train/eval horizons.
   
   This setting is modeled off of v1.1-v1.4 It is appropriate as a quick train
   task for new ideas, a transfer target for agents trained on large maps,
   or as a primary research target for PCG methods.'''

   #Memory/Batch Scale
   NUM_WORKERS             = 28
   TRAIN_BATCH_SIZE        = 64 * 128 * NUM_WORKERS
   SGD_MINIBATCH_SIZE      = 1 * 128
   ROLLOUT_FRAGMENT_LENGTH = 256
 
   #Horizon
   TRAIN_HORIZON           = 1024
   EVALUATION_HORIZON      = 1024


class Debug(SmallMaps, config.AllGameSystems):
   '''Debug Neural MMO training setting

   A version of the SmallMap setting with greatly reduced batch parameters.
   Only intended as a tool for identifying bugs in the model or environment'''
   LOAD                    = False
   LOCAL_MODE              = True
   NUM_WORKERS             = 1

   SGD_MINIBATCH_SIZE      = 100
   TRAIN_BATCH_SIZE        = 400
   TRAIN_HORIZON           = 200
   EVALUATION_HORIZON      = 50

   HIDDEN                  = 2
   EMBED                   = 2


### AICrowd competition settings
class CompetitionRound1(config.Achievement, SmallMaps):
   @property
   def SPAWN(self):
      return self.SPAWN_CONCURRENT

   NENT                    = 128
   NPOP                    = 1

class CompetitionRound2(config.Achievement, SmallMaps):
   @property
   def SPAWN(self):
      return self.SPAWN_CONCURRENT

   NENT                    = 128
   NPOP                    = 16
   COOPERATIVE             = True

class CompetitionRound3(config.Achievement, LargeMaps):
   @property
   def SPAWN(self):
      return self.SPAWN_CONCURRENT

   NENT                    = 1024
   NPOP                    = 32
   COOPERATIVE             = True


### NeurIPS Experiments
class SmallMultimodalSkills(SmallMaps, config.AllGameSystems): pass
class LargeMultimodalSkills(LargeMaps, config.AllGameSystems): pass


class MagnifyExploration(SmallMaps, config.Resource, config.Progression):
   pass
class Population4(MagnifyExploration):
   NENT  = 4
class Population32(MagnifyExploration):
   NENT  = 32
class Population256(MagnifyExploration):
   NENT  = 256


class DomainRandomization16384(SmallMaps, config.AllGameSystems):
   TERRAIN_TRAIN_MAPS=16384
class DomainRandomization256(SmallMaps, config.AllGameSystems):
   TERRAIN_TRAIN_MAPS=256
class DomainRandomization32(SmallMaps, config.AllGameSystems):
   TERRAIN_TRAIN_MAPS=32
class DomainRandomization1(SmallMaps, config.AllGameSystems):
   TERRAIN_TRAIN_MAPS=1


class TeamBased(MagnifyExploration, config.Combat):
   NENT                    = 128
   NPOP                    = 32
   COOPERATIVE             = True
   TEAM_SPIRIT             = 0.5

   @property
   def SPAWN(self):
      return self.SPAWN_CONCURRENT



from pdb import set_trace as T
import numpy as np

from neural_mmo.forge.blade.io import node

def bind(gameCls):
   def to(ioCls):
      @property
      def GAME_CLS():
         return gameCls

      ioCls.GAME_CLS = GAME_CLS
      #ioCls.GAME_CLS = gameCls
      return ioCls
   return to 

class Config(metaclass=node.IterableNameComparable):
   pass

class Stimulus(Config):
   def dict():
      return {k[0] : v for k, v in dict(Stimulus).items()}

   class Entity(Config):
      @staticmethod
      def N(config):
         return config.N_AGENT_OBS

      class Self(node.Discrete):
         def init(self, config):
            self.max = 1
            self.scale = 1.0

      class ID(node.Continuous):
         def init(self, config):
            self.min   = -np.inf
            self.scale = 0.001

      class AttackerID(node.Continuous):
         def init(self, config):
            self.min   = -np.inf
            self.scale = 0.001

      class Level(node.Continuous):
         def init(self, config):
            self.scale = 0.05

      class Population(node.Discrete):
         def init(self, config):
            self.min = -3 #NPC index
            self.max = config.NPOP - 1
            self.scale = 1.0

      class R(node.Discrete):
         def init(self, config):
            self.min = 0
            self.max = config.TERRAIN_SIZE - 1
            self.scale = 0.15

      class C(node.Discrete):
         def init(self, config):
            self.min = 0
            self.max = config.TERRAIN_SIZE - 1
            self.scale = 0.15

      #Historical stats
      class Damage(node.Continuous):
         def init(self, config):
            #This scale may eventually be too high
            self.val   = 0
            self.scale = 0.1

      class TimeAlive(node.Continuous):
         def init(self, config):
            self.val = 0
            self.scale = 0.01

      #Resources -- Redo the max/min scaling. You can't change these
      #after init without messing up the embeddings
      class Food(node.Continuous):
         def init(self, config):
            if config.game_system_enabled('Progression'):
               self.val = config.PROGRESSION_BASE_RESOURCE
               self.max = config.PROGRESSION_BASE_RESOURCE
            elif config.game_system_enabled('Resource'):
               self.val = config.RESOURCE_BASE_RESOURCE
               self.max = config.RESOURCE_BASE_RESOURCE
            else:
               self.val = 1
               self.max = 1
    
            self.scale = 0.1

      class Water(node.Continuous):
         def init(self, config):
            if config.game_system_enabled('Progression'):
               self.val = config.PROGRESSION_BASE_RESOURCE
               self.max = config.PROGRESSION_BASE_RESOURCE
            elif config.game_system_enabled('Resource'):
               self.val = config.RESOURCE_BASE_RESOURCE
               self.max = config.RESOURCE_BASE_RESOURCE
            else:
               self.val = 1
               self.max = 1
 
            self.scale = 0.1

      class Health(node.Continuous):
         def init(self, config):
            self.val = config.BASE_HEALTH
            self.max = config.BASE_HEALTH
            self.scale = 0.1

      #Status effects
      class Freeze(node.Continuous):
         def init(self, config):
            self.val = 0
            self.max = 3
            self.scale = 0.3

   class Tile(Config):
      @staticmethod
      def N(config):
         return config.WINDOW**2

      class NEnts(node.Continuous):
         def init(self, config):
            self.max = config.NENT
            self.val = 0
            self.scale = 1.0

      class Index(node.Discrete):
         def init(self, config):
            self.max = config.NTILE
            self.scale = 0.15

      class R(node.Discrete):
         def init(self, config):
            self.max = config.TERRAIN_SIZE - 1
            self.scale = 0.15
 
      class C(node.Discrete):
         def init(self, config):
            self.max = config.TERRAIN_SIZE - 1
            self.scale = 0.15

for objName, obj in Stimulus:
   for idx, (attrName, attr) in enumerate(obj):
      attr.index = idx 



from os.path import dirname, basename, isfile
import glob
def modules(f):
   modules = glob.glob(dirname(f)+"/*.py")
   return  [ basename(f)[:-3] for f in modules 
         if isfile(f) and not f.endswith('__init__.py')]





#Various Enums used for handling materials, entity types, etc.
#Data texture pairs are used for enums that require textures.
#These textures are filled in by the Render class at run time.

from pdb import set_trace as T
import numpy as np
import colorsys
from enum import Enum
#from forge.blade.item import ore
from neural_mmo.forge.blade import systems

class Defaults:
   BLACK    = (0, 0, 0)
   GRAY3    = (20, 20, 20)
   GRAY2    = (40, 40, 40)
   GRAY1    = (60, 60, 60)
   RED      = (255, 0, 0)
   GREEN    = (0, 255, 0)
   BLUE     = (0, 0, 255) 
   YELLOW   = (255, 255, 0)
   GOLD     = (212, 175, 55)
   MASK     = (214, 127, 255)

def rgb(h):
  h = h.lstrip('#')
  return tuple(int(h[i:i+2], 16) for i in (0, 2, 4))

def rgbNorm(h):
  h = h.lstrip('#')
  return tuple(int(h[i:i+2], 16)/255.0 for i in (0, 2, 4))

class Color:
    def __init__(self, name, hexVal):
        self.name = name
        self.hex = hexVal
        self.rgb = rgb(hexVal)
        self.norm = rgbNorm(hexVal)
        self.value = self.rgb #Emulate enum

    def packet(self):
        return self.hex

def makeColor(idx, h=1, s=1, v=1):
   r, g, b = colorsys.hsv_to_rgb(h, s, v)
   rgbval = tuple(int(255*e) for e in [r, g, b])
   hexval = '%02x%02x%02x' % rgbval
   return Color(str(idx), hexval)

class Color256:
   def make256():
      parh, parv = np.meshgrid(np.linspace(0.075, 1, 16), np.linspace(0.25, 1, 16)[::-1])
      parh, parv = parh.T.ravel(), parv.T.ravel()
      idxs = np.arange(256)
      params = zip(idxs, parh, parv)
      colors = [makeColor(idx, h=h, s=1, v=v) for idx, h, v in params]
      return colors
   colors = make256()

class Tier:
   BLACK    = Color('BLACK', '#000000')
   WOOD     = Color('WOOD', '#784d1d')
   BRONZE   = Color('BRONZE', '#db4508')
   SILVER   = Color('SILVER', '#dedede')
   GOLD     = Color('GOLD', '#ffae00')
   PLATINUM = Color('PLATINUM', '#cd75ff')
   DIAMOND  = Color('DIAMOND', '#00bbbb')
   

class Neon:
   RED      = Color('RED', '#ff0000')
   ORANGE   = Color('ORANGE', '#ff8000')
   YELLOW   = Color('YELLOW', '#ffff00')

   GREEN    = Color('GREEN', '#00ff00')
   MINT     = Color('MINT', '#00ff80')
   CYAN     = Color('CYAN', '#00ffff')

   BLUE     = Color('BLUE', '#0000ff')
   PURPLE   = Color('PURPLE', '#8000ff')
   MAGENTA  = Color('MAGENTA', '#ff00ff')

   FUCHSIA  = Color('FUCHSIA', '#ff0080')
   SPRING   = Color('SPRING', '#80ff80')
   SKY      = Color('SKY', '#0080ff')
 
   WHITE    = Color('WHITE', '#ffffff')
   GRAY     = Color('GRAY', '#666666')
   BLACK    = Color('BLACK', '#000000')

   BLOOD    = Color('BLOOD', '#bb0000')
   BROWN    = Color('BROWN', '#7a3402')
   GOLD     = Color('GOLD', '#eec600')
   SILVER   = Color('SILVER', '#b8b8b8')

   TERM     = Color('TERM', '#41ff00')
   MASK     = Color('MASK', '#d67fff')

   def color12():
      return (
              Neon.CYAN, Neon.MINT, Neon.GREEN,
              Neon.BLUE, Neon.PURPLE, Neon.MAGENTA,
              Neon.FUCHSIA, Neon.SPRING, Neon.SKY,
              Neon.RED, Neon.ORANGE, Neon.YELLOW)

   def rand12():
      twelveColor = color12()
      randInd = np.random.randint(0, len(twelveColor))
      return twelveColor[randInd]

class Solid:
   BLUE       = Color('BLUE', '#1f77b4')
   ORANGE     = Color('ORANGE', '#ff7f0e')
   GREEN      = Color('GREEN', '#2ca02c')

   RED        = Color('RED',  '#D62728')
   PURPLE     = Color('PURPLE', '#9467bd')
   BROWN      = Color('BROWN', '#8c564b')

   PINK       = Color('PINK', '#e377c2')
   GREY       = Color('GREY', '#7f7f7f')
   CHARTREUSE = Color('CHARTREUSE', '#bcbd22')

   SKY        = Color('SKY', '#17becf')

   def color10():
      return (
              Solid.BLUE, Solid.ORANGE, Solid.GREEN,
              Solid.RED, Solid.PURPLE, Solid.BROWN,
              Solid.PINK, Solid.CHARTREUSE, Solid.SKY,
              Solid.GREY)


class Palette:
   def __init__(self, n):
      self.n = n
      if n <= 12:
         self.colors = Neon.color12()
      else:
         self.colors = Color256.colors

   def color(self, idx):
      if self.n > 12:
         idx = int(idx * 256 // self.n)
      return self.colors[idx]

class DataTexturePair:
   def __init__(self, data, texture=None):
      self.data = data
      self.tex  = texture

   def __eq__(self, ind):
      return ind == self.data

   def __hash__(self):
      return hash(self.data)


class Entity(Enum):
   NONE     = DataTexturePair(0)
   NEURAL   = DataTexturePair(1)
   CHICKEN  = DataTexturePair(2)
   GOBLIN   = DataTexturePair(3)
   BEAR     = DataTexturePair(4)




from pdb import set_trace as T
import numpy as np
import time

from collections import defaultdict, deque

from neural_mmo.forge.blade.io.stimulus.static import Stimulus as Static
from neural_mmo.forge.blade.entity import Player

def camel(string):
   '''Convert a string to camel case'''
   return string[0].lower() + string[1:]

class Stimulus:
   '''Static IO class used for interacting with game observations

   The environment returns game objects in observations.
   This class assembles them into usable data packets'''
   def process(config, env, ent):
      '''Utility for preprocessing game observations

      Built to be semi-automatic and only require small updates
      for large new classes of observations

      Args:
         config    : An environment configuration object
         env       : Local environment observation
         ent       : Local entity observation
      '''

      #Static handles
      functions = [Stimulus.tile, Stimulus.entity]
      static    = [Static.Tile, Static.Entity]

      index = 0
      stim  = {}
      raw   = {}
      ent.stim = defaultdict(list)
      for func, stat in zip(functions, static):
         stim[stat], raw[stat] = func(env, ent, stat, config)

      return stim, raw

   def add(static, obj, config, args):
      '''Pull attributes from game'''
      #Cache observation representation
      if obj.repr is None:
          obj.repr = {}
          for name, attr in static:
             val = obj
             for n in name:
                val = val.__dict__[camel(n)]

             obj.repr[attr] = attr(config)

      stim = {}
      for name, attr in obj.repr.items():
         stim[name] = attr.get(*args)

      return stim

   def nop(template):
      stim = {}
      for attr in template:
         stim[attr] = np.zeros(1)
      return stim

   def tile(env, ent, static, config):
      '''Internal processor for tile objects'''
      stim = []
      raw = []
      for r, row in enumerate(env):
         for c, tile in enumerate(row):
            raw.append(tile)
            args = [tile, r, c]
            s = Stimulus.add(static, tile, config, args)
            s = dict(sorted(s.items()))
            ent.stim[type(tile)].append(tile)
            stim.append(s)
      return stim, raw

   def entity(env, ent, static, config):
      '''Internal processor for player objects. Always returns self first'''
      raw = []
      stim = deque()
      for tile in env.ravel():
         for e in tile.ents.values():
            if type(e) != Player:
               continue
            raw.append(e)
            args = [ent, e]
            s = Stimulus.add(static, e, config, args)
            s = dict(sorted(s.items()))

            ent.stim[type(e)].append(e)
            if ent is e:
               stim.appendleft(s)
            else:
               stim.append(s)

      #Should sort before truncate
      #Remember to also sort raw_obs
      stim = list(stim)[:config.N_AGENT_OBS]
      return stim, raw



from pdb import set_trace as T
import numpy as np
import torch
import time

from neural_mmo.forge.blade.lib.utils import EDA
import os

class Resetter:
   '''Utility for model stalling that keeps track
   of the time since the model has improved

   Args:
      resetTol: ticks before resetting
   '''
   def __init__(self, resetTol):
      self.resetTicks, self.resetTol = 0, resetTol 

   def step(self, best=False):
      '''Step the resetter

      Args:
         best (bool): whether this is the best model
            performance thus far

      Returns:
         bool: whether the model should be reset to
            the previous best checkpoint
      '''
      if best:
         self.resetTicks = 0
      elif self.resetTicks < self.resetTol:
         self.resetTicks += 1
      else:
         self.resetTicks = 0
         return True
      return False

class Saver:
   '''Model save/load class
   
   Args:
      root: Path for save
      savef: Name of the save file
      bestf: Name of the best checkpoint file
      resetTol: How often to reset if training stalls
   '''
   def __init__(self, root, savef, bestf, resetTol):
      self.bestf, self.savef = bestf, savef,
      self.root, self.extn = root, '.pth'
      
      self.resetter = Resetter(resetTol)
      self.rewardAvg, self.best = EDA(), 0
      self.start, self.epoch = time.time(), 0
      self.resetTol = resetTol

   def save(self, params, opt, fname):
      data = {'param': params,
              'opt' : opt,
              'epoch': self.epoch}
      path = os.path.join(self.root, fname + self.extn)
      torch.save(data, path)

   def checkpoint(self, params, opt, lifetime):
      '''Save the model to file

      Args:
         params: Parameters to save
         opt: Optimizer to save
         fname: File to save to
      '''
      #self.save(params, opt, self.savef)
      best = lifetime > self.best
      if best: 
         self.best = lifetime
         self.save(params, opt, self.bestf)

      self.time     = time.time() - self.start
      self.start    = time.time()
      self.lifetime = lifetime
      self.epoch += 1

      if self.epoch % 100 == 0:
         self.save(params, opt, 'model'+str(self.epoch))

      return self.perf(), self.resetter.step(best)

   def load(self, opt, param, best=False):
      '''Load the model from file

      Args:
         opt: Optimizer to load
         params: Parameters to load 
         best: Whether to load the best or latest checkpoint

      Returns:
         epoch: The epoch of the loaded checkpoint
      '''
      fname = self.bestf if best else self.savef
      path  = os.path.join(self.root, fname) + self.extn
      data  = torch.load(path)
      param.data = data['param']
      if opt is not None:
         opt.load(data['opt'])
      epoch = data['epoch']
      return epoch

   def perf(self):
      '''Print stats for the latest epoch'''
      return ''.join([
            'Tick: ', str(self.epoch),
            ', Time: ', str(self.time)[:5],
            ', Lifetime: ', str(self.lifetime)[:5],
            ', Best: ', str(self.best)[:5]])




from pdb import set_trace as T

import time
import ray

from functools import partial
from threading import Thread
from random import random

from bokeh.server.server import Server
from bokeh.models import ColumnDataSource
from bokeh.models import Band
from bokeh.plotting import curdoc, figure, show
from bokeh.themes import Theme

from tornado import gen
from tornado.ioloop import IOLoop

from neural_mmo.forge.blade.core.market.config import *

# to run a demo with dummy data:
# $ bokeh serve --show visualizer.py

# market plans
# click on objects in market to display stats about them
# market overview tab
# -> trade bandwidth, demand, gdp

PORT=5009

class MarketVisualizer:
    def __init__(self, keys, history_len: int = 512,
                 title: str = "NeuralMMO Market Data", x: str = "tick",
                 ylabel: str = "Dummy Values"):
        """Visualizes a stream of data with threaded refreshing
    
        Args:
            keys        : List of object names (str) to be displayed on market
            history_len : How far back to plot data
            title       : Title of graph
            x           : Name of x axis data
            ylabel      : Name of y axis on plot
        """
        self.colors = 'blue red green yellow orange purple'.split()
        self.data        = {}

        self.history_len = history_len
        self.title       = title
        self.keys        = keys

        self.ylabel      = ylabel
        self.x           = x

        # TODO figure out a better way to ensure unique colors
        assert len(self.keys) <= len(self.colors), 'Limited color pool'

        for i, key in enumerate(self.keys):
            self.data[key] = []
            self.data[key+'lower'] = []
            self.data[key+'upper'] = []
        self.data['tick'] = []


    def init(self, doc):
        #Source must only be modified through a stream
        self.source = ColumnDataSource(data=self.data)

        #Enable theming
        theme = Theme('forge/blade/core/market/theme.yaml')
        doc.theme = theme
        self.doc = doc

        fig = figure(
           plot_width=600,
           plot_height=400,
           tools='pan,xwheel_zoom,box_zoom,save,reset',
           title='Neural MMO: Market Data',
           x_axis_label=self.x,
           y_axis_label=self.ylabel)

        #Initialize plots
        for i, key in enumerate(self.keys):
            fig.line(
                source=self.source,
                x=self.x,
                y=key,
                color=self.colors[i],
                line_width=LINE_WIDTH,
                legend_label=key)

            band = Band(
               source=self.source,
               base=self.x,
               lower=key+'lower',
               upper=key+'upper',
               level='underlay',
               line_color=self.colors[i],
               line_width=1,
               line_alpha=0.2,
               fill_color=self.colors[i],
               fill_alpha=0.2)
            fig.add_layout(band) 

        #Set root
        self.doc.add_root(fig)
        self.fig = fig

@ray.remote
class BokehServer:
    def __init__(self, market, *args, **kwargs):
        """ Runs an asynchronous Bokeh data streaming server
      
        Args:
           market : The market to visualize
           args   : Additional arguments
           kwargs : Additional keyword arguments
        """
        self.visu   = MarketVisualizer(*args, **kwargs)
        self.market = market

        server = Server(
                {'/': self.init},
                io_loop=IOLoop.current(),
                port=PORT,
                num_procs=1)

        self.server = server
        self.thread = None
        self.tick   = 0
        server.start()

        server.io_loop.add_callback(server.show, "/")
        server.io_loop.start()

    def init(self, doc):
        '''Initialize document and threaded update loop
      
        Args:
           doc: A Bokeh document
        '''
        self.visu.init(doc)
        self.doc = doc

        self.thread = Thread(target=self.update, args=[])
        self.thread.start()
        self.started = True

    def update(self):
        '''Blocking update call to be run in a separate thread
        
        Ingests packets from a remote market and streams to Bokeh client'''
        self.n = 0
        while True:
            #Wait for thread to initialize
            time.sleep(0.05)
            if self.thread is None:
               continue 

            #Get remote market data
            packet = ray.get(self.market.getData.remote())
            if packet is None:
               continue

            #Ingest market data
            for key, val in packet.items():
              if key[-3:] == 'std':
                 key = key[:-4]
                 dat = packet[key]
                 self.visu.data[key + 'lower'].append(dat - val)
                 self.visu.data[key + 'upper'].append(dat + val)
              else:
                 self.visu.data[key].append(val)

            #Stream to Bokeh client
            self.doc.add_next_tick_callback(partial(self.stream))
            self.tick += 1

    @gen.coroutine
    def stream(self):
        '''Stream current data buffer to Bokeh client'''
        self.visu.source.stream(self.visu.data, self.visu.history_len)

@ray.remote
class Middleman:
    def __init__(self):
        '''Remote data buffer for two processes to dump and recv data

        This is probably not safe'''
        self.data = None
        
    def getData(self):
        '''Get data from buffer

        Returns:
           data: From buffer
        '''
        data = self.data
        self.data = None
        return data

    def setData(self, data):
        '''Set buffer data

        Args:
           data: To set buffer
        '''
        self.data = data

class Market:
    def __init__(self, items, middleman):
        '''Dummy market emulator

        Args: 
           items     : List of item keys
           middleman : A Middleman object'''
        self.middleman = middleman
        self.items     = items

        self.keys = items
        self.data = {}
        self.tick = 0

        self.data['tick'] = 0
        for i, key in enumerate(self.keys):
            self.data[key] = 0

    def update(self):
        '''Updates market data and propagates to Bokeh server

        Note: best to update all at once. Current version may cause bugs'''
        for key, val in self.data.items():
            self.data[key] = val + 0.2*(random() - 0.5)
            if key == 'tick':
                self.data[key] = self.tick

        self.tick += 1
        self.middleman.setData.remote(self.data)

# Example setup
if __name__ == '__main__':
   ray.init()
   ITEMS = ['Food', 'Water', 'Health', 'Melee', 'Range', 'Mage']

   middleman  = Middleman.remote()
   market     = Market(ITEMS, middleman)
   visualizer = BokehServer.remote(middleman, ITEMS)

   while True:
     time.sleep(0.1)
     market.update()




from neural_mmo.forge.blade.item import item, armor

class Equipment:
   def __init__(self):
      self.ammo = 0
      self.resetArmor()
      self.resetMelee()
      self.resetRanged()

   def resetArmor(self):
      self.armor = armor.Base()

   def resetMelee(self):
      self.melee = item.Base()
      
   def resetRanged(self):
      self.ranged = item.Base()




from neural_mmo.forge.blade.systems.ai import utils
from neural_mmo.forge.blade.systems.ai import move
from neural_mmo.forge.blade.systems.ai import attack
from neural_mmo.forge.blade.systems.ai import behavior
from neural_mmo.forge.blade.systems.ai import policy



'''Observation processing module'''

import torch
from torch import nn

from neural_mmo.forge.blade.io import stimulus


class Input(nn.Module):
   def __init__(self, config, embeddings, attributes):
      '''Network responsible for processing observations

      Args:
         config     : A configuration object
         embeddings : An attribute embedding module
         attributes : An attribute attention module
         entities   : An entity attention module
      '''
      super().__init__()

      self.embeddings = nn.ModuleDict()
      self.attributes = nn.ModuleDict()

      for _, entity in stimulus.Static:
         continuous = len([e for e in entity if e[1].CONTINUOUS])
         discrete   = len([e for e in entity if e[1].DISCRETE])
         #self.attributes[entity.__name__] = attributes(config.EMBED, config.HIDDEN)
         self.attributes[entity.__name__] = nn.Linear(
               (continuous+discrete)*config.HIDDEN, config.HIDDEN)
         self.embeddings[entity.__name__] = embeddings(
               continuous=continuous, discrete=4096, config=config)

      
      #TODO: implement obs scaling in a less hackey place
      self.tileWeight = torch.Tensor([1.0, 0.0, 0.02, 0.02])
      self.entWeight  = torch.Tensor([1.0, 0.0, 0.0, 0.05, 0.00, 0.02, 0.02, 0.1, 0.01, 0.1, 0.1, 0.1, 0.3])
      if torch.cuda.is_available():
         self.tileWeight = self.tileWeight.cuda()
         self.entWeight  = self.entWeight.cuda()

   def forward(self, inp):
      '''Produces tensor representations from an IO object

      Args:                                                                   
         inp: An IO object specifying observations                      
         

      Returns:
         observationTensor : A fixed size observation representation
         entityLookup      : A fixed size representation of each entity
      ''' 
      #Pack entities of each attribute set
      entityLookup = {}

      inp['Tile']['Continuous']   *= self.tileWeight
      inp['Entity']['Continuous'] *= self.entWeight
 
      entityLookup['N'] = inp['Entity'].pop('N')
      for name, entities in inp.items():
         #Construct: Batch, ents, nattrs, hidden
         embeddings = self.embeddings[name](entities)
         B, N, _, _ = embeddings.shape
         embeddings = embeddings.view(B, N, -1)

         #Construct: Batch, ents, hidden
         entityLookup[name] = self.attributes[name](embeddings)

      return entityLookup



from pdb import set_trace as T

import scipy.stats as stats
from matplotlib import pyplot as plt
import numpy as np

import os

import vec_noise
from imageio import imread, imsave
from tqdm import tqdm

from neural_mmo.forge.blade.lib import material

def mkdir(path):
   try:
      os.mkdir(path)
   except:
      pass

def sharp(self, noise):
   return 2 * (0.5 - abs(0.5 - noise));

class Save:
   def render(mats, lookup, path):
      images = [[lookup[e] for e in l] for l in mats]
      image = np.vstack([np.hstack(e) for e in images])
      imsave(path, image)

   def fractal(terrain, path):
      frac = (256*terrain).astype(np.uint8)
      imsave(path, frac)

   def np(mats, path):
      '''Saves a map into into a tiled compatiable file given a save_path, width
       and height of the map, and 2D numpy array specifiying enums for the array'''
      mkdir(path)
      path = os.path.join(path, 'map.npy')
      np.save(path, mats.astype(np.int))

class Terrain:
   pass

class MapGenerator:
   def __init__(self, config):
      self.config = config
      self.loadTextures()

   def loadTextures(self):
      lookup = {}
      path = self.config.PATH_TILE
      for mat in material.All:
         key = mat.tex
         tex = imread(path.format(key))
         mat.tex = tex[:, :, :3][::4, ::4]
         mat.tex = mat.tex.reshape(-1, 3).mean(0).astype(np.uint8)
         lookup[mat.index] = mat.tex.reshape(1, 1, 3)
         #lookup[mat.index] = mat.tex
         setattr(Terrain, key.upper(), mat.index)
      self.textures = lookup

   def material(self, config, val):
      if val <= config.TERRAIN_WATER:
         return Terrain.WATER
      if val <= config.TERRAIN_GRASS:
         return Terrain.GRASS
      if val <= config.TERRAIN_FOREST:
         return Terrain.FOREST
      return Terrain.STONE

   def generate(self):
      config = self.config
      if config.__class__.__name__ == 'SmallMaps':
         prefix = config.PATH_MAPS_SMALL
      elif config.__class__.__name__ == 'LargeMaps':
         prefix = config.PATH_MAPS_LARGE
      else:
         prefix = config.PATH_MAPS

      #Train and eval map indices
      msg    = 'Generating {} training and {} evaluation maps:'
      evalMaps  = range(-config.TERRAIN_EVAL_MAPS, 0)
      trainMaps = range(1, config.TERRAIN_TRAIN_MAPS+1)
      print(msg.format(config.TERRAIN_TRAIN_MAPS, config.TERRAIN_EVAL_MAPS))

      llen = config.TERRAIN_EVAL_MAPS + config.TERRAIN_TRAIN_MAPS
      perm = np.random.RandomState(seed=0).permutation(llen)
      interpolaters = np.logspace(config.TERRAIN_LOG_INTERPOLATE_MIN,
                                  config.TERRAIN_LOG_INTERPOLATE_MAX,
                                  llen)[perm]

      for idx, seed in enumerate(tqdm([*evalMaps, *trainMaps])):
         path = prefix + '/map' + str(seed)
         mkdir(prefix)
         mkdir(path)

         terrain, tiles = self.grid(config, seed, interpolaters[idx])

         Save.np(tiles, path)
         if config.TERRAIN_RENDER:
            Save.fractal(terrain, path+'/fractal.png')
            Save.render(tiles, self.textures, path+'/map.png')

   def grid(self, config, seed, interpolate):
      center      = config.TERRAIN_CENTER
      border      = config.TERRAIN_BORDER
      size        = config.TERRAIN_SIZE
      frequency   = config.TERRAIN_FREQUENCY
      offset      = config.TERRAIN_FREQUENCY_OFFSET
      octaves     = center // config.TERRAIN_TILES_PER_OCTAVE

      #Data buffers
      val   = np.zeros((size, size, octaves))
      scale = np.zeros((size, size, octaves))
      s     = np.arange(size)
      X, Y  = np.meshgrid(s, s)

      #Compute noise over logscaled octaves
      start = frequency
      end   = min(start, start - np.log2(center) + offset)
      for idx, freq in enumerate(np.logspace(start, end, octaves, base=2)):
         val[:, :, idx] = vec_noise.snoise2(seed*size + freq*X, idx*size + freq*Y)

      #Compute L1 distance
      x      = np.abs(np.arange(size) - size//2)
      X, Y   = np.meshgrid(x, x)
      data   = np.stack((X, Y), -1)
      l1     = np.max(abs(data), -1)

      #Interpolation Weights
      rrange = np.linspace(-1, 1, 2*octaves-1)
      pdf    = stats.norm.pdf(rrange, 0, interpolate)
      pdf    = pdf / max(pdf)
      high   = center / 2
      delta  = high / octaves

      #Compute perlin mask
      noise  = np.zeros((size, size))
      X, Y   = np.meshgrid(s, s)
      expand = int(np.log2(center)) - 2
      for idx, octave in enumerate(range(expand, 1, -1)):
         freq, mag = 1 / 2**octave, 1 / 2**idx
         noise    += mag * vec_noise.snoise2(seed*size + freq*X, idx*size + freq*Y) 

      #plt.imshow(noise)
      #plt.show()
      noise -= np.min(noise)
      noise = octaves * noise / np.max(noise) - 1e-12
      noise = noise.astype(np.int)
      #plt.imshow(noise)
      #plt.show()

      #Compute L1 and Perlin scale factor
      for i in range(octaves):
         start             = octaves - i - 1
         scale[l1 <= high] = np.arange(start, start + octaves)
         high             -= delta

      start   = noise - 1
      l1Scale = np.clip(l1, 0, size//2 - border - 2) 
      l1Scale = l1Scale / np.max(l1Scale)
      for i in range(octaves):
         idxs           = l1Scale*scale[:, :, i] + (1-l1Scale)*(start + i)
         scale[:, :, i] = pdf[idxs.astype(np.int)]

      #Blend octaves
      std = np.std(val)
      val = val / std
      val = scale * val
      val = np.sum(scale * val, -1)
      val = std * val / np.std(val)
      val = 0.5 + np.clip(val, -1, 1)/2

      #Threshold to materials
      matl = np.zeros((size, size), dtype=object)
      for y in range(size):
         for x in range(size):
            matl[y, x] = self.material(config, val[y, x])

      #Lava and grass border
      matl[l1 > size/2 - border]       = Terrain.LAVA
      matl[l1 == size//2 - border]     = Terrain.GRASS

      edge  = l1 == size//2 - border - 1
      stone = (matl == Terrain.STONE) | (matl == Terrain.WATER)
      matl[edge & stone] = Terrain.FOREST

      return val, matl



def isInt(x):
   return type(x) in (float, int) and int(x) == x

class CommChannel:
   def __init__(self):
      self.outbox = []
      self.inbox = []

   def put(self, update):
      if len(update) > 0:
         self.outbox.append(update)

   def get(self):
      inbox = self.inbox
      self.inbox = []
      return inbox

   def send(self):
      outbox = self.outbox
      self.outbox = []
      return outbox

   def recv(self, updates):
      self.inbox += updates



from pdb import set_trace as T
import numpy as np
import torch
import time

from collections import defaultdict
from torch.nn.parameter import Parameter

from neural_mmo.forge.ethyr.torch import save
from neural_mmo.forge.ethyr.torch.optim import ManualAdam
from neural_mmo.forge.ethyr.torch.param import setParameters, getParameters

class GradientOptimizer:
   def __init__(self, net, config):
      self.config = config
      self.net    = net

      self.opt = ManualAdam([net.parameters], 
         lr=self.config.LR, weight_decay=self.config.DECAY)

   #Grads and clip
   def step(self, grads, logs):
      '''Clip the provided gradients and step the optimizer

      Args:
         gradList: a list of gradients
      '''
      grads = np.mean(grads, 0)
      print('Gradient magnitude: ', np.sqrt(np.sum(grads**2)))
      mag = self.config.GRAD_CLIP
      grads = np.clip(grads, -mag, mag)

      gradTensor = torch.Tensor(grads)
      self.opt.step(gradTensor)

      self.net.syncParameters()

   def load(self, opt):
      self.opt.load_state_dict(opt.opt.state_dict())
      return self

class PopulationOptimizer:
   def __init__(self, model, config):
      self.config = config
      self.model  = model 

   def step(self, gradList, logs):
      lifetimes = defaultdict(list)
      for blob in logs.blobs:
         lifetimes[blob.annID].append(blob.lifetime)

      performance = {}
      for key, val in lifetimes.items():
         performance[key] = np.mean(val)

      performance = sorted(performance.items(), 
         key=lambda x: x[1], reverse=True)
      
      for idx in range(self.config.PERMPOPS):
         goodIdx, goodPerf = performance[idx]
         badIdx,  badPerf  = performance[-idx-1]

         self.permuteNet(goodIdx, badIdx)

   def permuteNet(self, goodIdx, badIdx):
      goodNet = self.model.net.net[goodIdx]
      badNet  = self.model.net.net[badIdx]

      goodParams = getParameters(goodNet)
      noise      = self.config.PERMVAL * np.random.randn(len(goodParams))
      goodParams = np.array(goodParams) + noise
      setParameters(badNet, goodParams)

class Model:
   '''Model manager class

   Convenience class wrapping saving/loading,
   model initialization, optimization, and logging.

   Args:
      ann: Model to optimize. Used to initialize weights.
      config: A Config specification
      args: Hook for additional user arguments
   '''
   def __init__(self, ann, config):
      self.saver = save.Saver(config.MODELDIR,
            'models', 'bests', resetTol=256)
      self.config = config

      print('Initializing new model...')
      self.net = ann(config)
      self.parameters = Parameter(torch.Tensor(
            np.array(getParameters(self.net))))

      #Have been experimenting with population based
      #training. Nothing stable yet -- advise avoiding
      if config.POPOPT:
         self.opt = PopulationOptimizer(self, config)
      else:
         self.opt = GradientOptimizer(self, config)

      if config.LOAD or config.BEST:
         self.load(self.opt, config.BEST)

   def step(self, recvs, blobs, log, lifetime):
      if self.config.TEST:
         return

      self.opt.step(recvs, blobs)
      perf, _ = self.checkpoint(self.opt, lifetime)
      return perf

   def load(self, opt, best=False):
      '''Load a model from file

      Args:
         best (bool): Whether to load the best (True)
             or most recent (False) checkpoint
      '''
      print('Loading model...')
      epoch = self.saver.load(opt, self.parameters, best)
      self.syncParameters()
      return self

   def checkpoint(self, opt, lifetime):
      '''Save the model to checkpoint

      Args:
         reward: Mean reward of the model
      '''
      return self.saver.checkpoint(self.parameters, opt, lifetime)

   def printParams(self):
      '''Print the number of model parameters'''
      nParams = len(self.weights)
      print('#Params: ', str(nParams/1000), 'K')

   def syncParameters(self):
      parameters = self.parameters.detach().numpy().tolist()
      setParameters(self.net, parameters)

   @property
   def weights(self):
      '''Get model parameters

      Returns:
         a numpy array of model parameters
      '''
      return getParameters(self.net)





from pdb import set_trace as T

from neural_mmo.forge.trinity.agent import Agent
from neural_mmo.forge.trinity.scripted import behavior, move, attack, utils, io
from neural_mmo.forge.blade.io.stimulus.static import Stimulus
from neural_mmo.forge.blade.io.action import static as Action
from neural_mmo.forge.blade.lib import enums

class Scripted(Agent):
    '''Template class for scripted models.

    You may either subclass directly or mirror the __call__ function'''
    scripted = True
    color    = enums.Neon.SKY
    def __init__(self, config, idx):
        '''
        Args:
           config : A forge.blade.core.Config object or subclass object
        ''' 
        super().__init__(config, idx)
        self.food_max  = 0
        self.water_max = 0

        self.spawnR    = None
        self.spawnC    = None

    @property
    def forage_criterion(self) -> bool:
        '''Return true if low on food or water'''
        min_level = 7
        return self.food <= min_level or self.water <= min_level

    def forage(self):
        '''Min/max food and water using Dijkstra's algorithm'''
        move.forageDijkstra(self.config, self.ob, self.actions, self.food_max, self.water_max)

    def explore(self):
        '''Route away from spawn'''
        move.explore(self.config, self.ob, self.actions, self.spawnR, self.spawnC)

    @property
    def downtime(self):
        '''Return true if agent is not occupied with a high-priority action'''
        return not self.forage_criterion and self.attacker is None

    def evade(self):
        '''Target and path away from an attacker'''
        move.evade(self.config, self.ob, self.actions, self.attacker)
        self.target     = self.attacker
        self.targetID   = self.attackerID
        self.targetDist = self.attackerDist

    def attack(self):
        '''Attack the current target'''
        if self.target is not None:
           assert self.targetID is not None
           attack.target(self.config, self.actions, self.style, self.targetID)

    def select_combat_style(self):
       '''Select a combat style based on distance from the current target'''
       if self.target is None:
          return

       if self.targetDist <= self.config.COMBAT_MELEE_REACH:
          self.style = Action.Melee
       elif self.targetDist <= self.config.COMBAT_RANGE_REACH:
          self.style = Action.Range
       else:
          self.style = Action.Mage

    def target_weak(self):
        '''Target the nearest agent if it is weak'''
        if self.closest is None:
            return False

        selfLevel = io.Observation.attribute(self.ob.agent, Stimulus.Entity.Level)
        targLevel = io.Observation.attribute(self.closest, Stimulus.Entity.Level)
        
        if targLevel <= selfLevel <= 5 or selfLevel >= targLevel + 3:
           self.target     = self.closest
           self.targetID   = self.closestID
           self.targetDist = self.closestDist

    def scan_agents(self):
        '''Scan the nearby area for agents'''
        self.closest, self.closestDist   = attack.closestTarget(self.config, self.ob)
        self.attacker, self.attackerDist = attack.attacker(self.config, self.ob)

        self.closestID = None
        if self.closest is not None:
           self.closestID = io.Observation.attribute(self.closest, Stimulus.Entity.ID)

        self.attackerID = None
        if self.attacker is not None:
           self.attackerID = io.Observation.attribute(self.attacker, Stimulus.Entity.ID)

        self.style      = None
        self.target     = None
        self.targetID   = None
        self.targetDist = None

    def adaptive_control_and_targeting(self, explore=True):
        '''Balanced foraging, evasion, and exploration'''
        self.scan_agents()

        if self.attacker is not None:
           self.evade()
           return

        if self.forage_criterion or not explore:
           self.forage()
        else:
           self.explore()

        self.target_weak()

    def __call__(self, obs):
        '''Process observations and return actions

        Args:
           obs: An observation object from the environment. Unpack with io.Observation
        '''
        self.actions = {}

        self.ob = io.Observation(self.config, obs)
        agent   = self.ob.agent

        self.food   = io.Observation.attribute(agent, Stimulus.Entity.Food)
        self.water  = io.Observation.attribute(agent, Stimulus.Entity.Water)

        if self.food > self.food_max:
           self.food_max = self.food
        if self.water > self.water_max:
           self.water_max = self.water

        if self.spawnR is None:
            self.spawnR = io.Observation.attribute(agent, Stimulus.Entity.R)
        if self.spawnC is None:
            self.spawnC = io.Observation.attribute(agent, Stimulus.Entity.C)

class Random(Scripted):
    name = 'Random_'
    '''Moves randomly'''
    def __call__(self, obs):
        super().__call__(obs)

        move.random(self.config, self.ob, self.actions)
        return self.actions

class Meander(Scripted):
    name = 'Meander_'
    '''Moves randomly on safe terrain'''
    def __call__(self, obs):
        super().__call__(obs)

        move.meander(self.config, self.ob, self.actions)
        return self.actions

class ForageNoExplore(Scripted):
    '''Forages using Dijkstra's algorithm'''
    name = 'ForageNE_'
    def __call__(self, obs):
        super().__call__(obs)

        self.forage()

        return self.actions

class Forage(Scripted):
    '''Forages using Dijkstra's algorithm and actively explores'''
    name = 'Forage_'
    def __call__(self, obs):
        super().__call__(obs)

        if self.forage_criterion:
           self.forage()
        else:
           self.explore()

        return self.actions

class CombatNoExplore(Scripted):
    '''Forages using Dijkstra's algorithm and fights nearby agents'''
    name = 'CombatNE_'
    def __call__(self, obs):
        super().__call__(obs)

        self.adaptive_control_and_targeting(explore=False)

        self.style = Action.Range
        self.attack()

        return self.actions
 
class Combat(Scripted):
    '''Forages, fights, and explores'''
    name = 'Combat_'
    def __call__(self, obs):
        super().__call__(obs)

        self.adaptive_control_and_targeting()

        self.style = Action.Range
        self.attack()

        return self.actions

class CombatTribrid(Scripted):
    name = 'CombatTri_'
    '''Forages, fights, and explores.

    Uses a slightly more sophisticated attack routine'''
    def __call__(self, obs):
        super().__call__(obs)

        self.adaptive_control_and_targeting()

        self.select_combat_style()
        self.attack()

        return self.actions



from neural_mmo.forge.blade.lib.multiset import MultiSet
from neural_mmo.forge.blade.lib.priorityqueue import PriorityQueue
from neural_mmo.forge.blade.lib.stattrak import StatTraker
from neural_mmo.forge.blade.lib.ray import init
from neural_mmo.forge.blade.lib.ray import clearbuffers



from pdb import set_trace as T

from neural_mmo.forge.blade.io.stimulus.static import Stimulus

def l1(start, goal):
   sr, sc = start
   gr, gc = goal
   return abs(gr - sr) + abs(gc - sc)

def l2(start, goal):
   sr, sc = start
   gr, gc = goal
   return 0.5*((gr - sr)**2 + (gc - sc)**2)**0.5

def lInfty(start, goal):
   sr, sc = start
   gr, gc = goal
   return max(abs(gr - sr), abs(gc - sc))

def adjacentPos(pos):
   r, c = pos
   return [(r - 1, c), (r, c - 1), (r + 1, c), (r, c + 1)]

def adjacentDeltas():
   return [(-1, 0), (1, 0), (0, 1), (0, -1)]

def inSight(dr, dc, vision):
    return (
          dr >= -vision and
          dc >= -vision and
          dr <= vision and
          dc <= vision)

def vacant(tile):
   Tile     = Stimulus.Tile
   occupied = Observation.attribute(tile, Tile.NEnts)
   matl     = Observation.attribute(tile, Tile.Index)

   lava    = material.Lava.index
   water   = material.Water.index
   grass   = material.Grass.index
   scrub   = material.Scrub.index
   forest  = material.Forest.index
   stone   = material.Stone.index
   orerock = material.Orerock.index

   return matl in (grass, scrub, forest) and not occupied



import itertools
import time

import numpy as np
from collections import defaultdict

class staticproperty(property):
    def __get__(self, cls, owner):
        return self.fget.__get__(None, owner)()

class classproperty(object):
    def __init__(self, f):
        self.f = f
    def __get__(self, obj, owner):
        return self.f(owner)

#https://stackoverflow.com/questions/26496831/how-to-convert-defaultdict-of-defaultdicts-of-defaultdicts-to-dict-of-dicts-o
def default_to_regular(d):
    if isinstance(d, defaultdict):
        d = {k: default_to_regular(v) for k, v in d.items()}
    return d

def cosSim(x):
   from sklearn.metrics.pairwise import euclidean_distances as pdist
   mag = np.sqrt(np.sum(x**2, 1))
   x = x / mag.reshape(-1, 1)
   dists = pdist(x)
   return dists

def vstack(x):
   if len(x) > 0:
      return np.vstack(x)
   return []

def groupby(items, key):
   targs = sorted(items, key=key)
   return itertools.groupby(targs, key=key)

#Generic
def uniqueKey(refDict):
   idx = seed()
   if idx in refDict:
      #print('Hash collision--highly unlikely, check your code')
      return uniqueKey(refDict)
   return idx

def seed():
   return int(np.random.randint(0, 2**32))

def invertDict(x):
   return {v: k for k, v in x.items()}

def loadDict(fName):
   with open(fName) as f:
      s = eval(f.read())
   return s

def terminalClasses(cls):
   ret = []
   subclasses = cls.__subclasses__()
   if len(subclasses) == 0:
      ret += [cls]
   else:
      for e in subclasses:
         ret += terminalClasses(e)
   return ret

def l1(pos1, pos2):
   r1, c1 = pos1
   r2, c2 = pos2
   return abs(r1 - r2) + abs(c1 - c2)

def l2(pos1, pos2):
   r1, c1 = pos1
   r2, c2 = pos2
   return np.sqrt((r1 - r2)**2 + (c1 - c2)**2)

def linf(pos1, pos2):
   r1, c1 = pos1
   r2, c2 = pos2
   return max(abs(r1 - r2), abs(c1 - c2))

def norm(x, n=2):
   return (np.sum(np.abs(x)**n)**(1.0/n)) / np.prod(x.shape)

#Bounds checker
def inBounds(r, c, shape, border=0):
   R, C = shape
   return (
         r > border and
         c > border and
         r < R - border and
         c < C - border
         )

#Because the numpy version is horrible
def randomChoice(aList):
   lLen = len(aList)
   ind = np.random.randint(0, lLen)
   return aList[ind]

#Tracks inds of a permutation
class Perm():
   def __init__(self, n):
      self.inds = np.random.permutation(np.arange(n))
      self.m = n
      self.pos = 0

   def next(self, n):
      assert(self.pos + n < self.m)
      ret = self.inds[self.pos:(self.pos+n)]
      self.pos += n
      return ret

#Exponentially decaying average
class EDA():
   def __init__(self, k=0.9):
      self.k = k 
      self.eda = None
   
   def update(self, x):
      if self.eda is None:
         self.eda = x
         return
      #self.eda = self.eda * k / (x * (1-k))
      self.eda = (1-self.k)*x + self.k*self.eda

class Timer:
   def __init__(self):
      self.start = time.time()

   def ticked(self, delta):
      if time.time() - self.start > delta:
         self.start = time.time()
         return True
      return False

class BenchmarkTimer:
   def __init__(self):
      self.eda = EDA() 
      self.accum = 0
   def startRecord(self):
      self.start = time.time()

   def stopRecord(self, accum=False):
      if accum:
         self.accum += time.time() - self.start
      else:
         self.eda.update(self.accum + time.time() - self.start)
         self.accum = 0

   def benchmark(self):
      return self.eda.eda
      
#Continuous moving average
class CMA():
   def __init__(self):
      self.t = 1.0
      self.cma = None

   def update(self, x):
      if self.cma is None:
         self.cma = x
         return
      self.cma = (x + self.t*self.cma)/(self.t+1)
      self.t += 1.0

#Continuous moving average
class CMV():
   def __init__(self):
      self.cma = CMA()
      self.cmv = None

   def update(self, x):
      if self.cmv is None:
         self.cma.update(x)
         self.cmv = 0
         return
      prevMean = self.cma.cma
      self.cma.update(x)
      self.cmv += (x-prevMean)*(x-self.cma.cma)

   @property
   def stats(self):
      return self.cma.cma, self.cmv

def matCrop(mat, pos, stimSz):
   ret = np.zeros((2*stimSz+1, 2*stimSz+1), dtype=np.int32)
   R, C = pos
   rt, rb = R-stimSz, R+stimSz+1
   cl, cr = C-stimSz, C+stimSz+1
   for r in range(rt, rb):
      for c in range(cl, cr):
         if inBounds(r, c, mat.shape):
            ret[r-rt, c-cl] = mat[r, c]
         else:
            ret[r-rt, c-cl] = 0
   return ret




from pdb import set_trace as T
import numpy as np

import torch
from torch import nn

from neural_mmo.forge.ethyr.torch import utils

class ReluBlock(nn.Module):
   def __init__(self, h, layers=2, postRelu=True):
      super().__init__()
      self.postRelu = postRelu

      self.layers = utils.ModuleList(
         nn.Linear, h, h, n=layers)

   def forward(self, x):
      for idx, fc in enumerate(self.layers):
         if idx != 0:
            x = torch.relu(x)
         x = fc(x)

      if self.postRelu:
         x = torch.relu(x)

      return x



#Various high level routines and tools for building quick
#NPC AIs. Returns only action arguments where possible. More
#complex routines return (action, args) pairs where required.

import numpy as np
from queue import Queue
from neural_mmo.forge.blade.lib import enums, utils
from neural_mmo.forge.blade.io.action import static as Action

# Flighty 
# PassiveAggressive
# Hostile 

def turfSearchAndDestroy(world, entity, whitelist):
   for e in sortedByl1(world.env.ent, world.size, entity):
     if e.entityIndex == enums.Entity.NEURAL.value.data:
         if isAdjacent(entity.pos, e.pos):
            tree = Actions.ActionTree(world, entity, rootVersion=Actions.MeleeV2) 
            tree.decideArgs(e)
            return tree.actionArgPair()
         move = routePath(entity.pos, e.pos)
         if inWhitelist(world.env.tiles, entity.pos, move, whitelist):
            tree = Actions.ActionTree(world, entity, rootVersion=Actions.MoveV2) 
            tree.decideArgs(move)
            return tree.actionArgPair()
   return randomOnTurf(world, entity, whitelist)

def randomMove(world, entity):
   tree = Actions.ActionTree(world, entity, rootVersion=Actions.MoveV2) 
   tree.randomArgs()
   return tree.actionArgPair()

def randomOnTurf(world, entity, whitelist):
   delMatPairs = adjacencyDelMatPairs(world.env, entity.pos)
   moves = whitelistByBlock(delMatPairs, whitelist)
   if len(moves) == 0: return Actions.Pass(), Actions.EmptyArgs()
   ind = np.random.randint(0, len(moves))

   tree = Actions.ActionTree(world, entity, rootVersion=Actions.MoveV2) 
   tree.decideArgs(moves[ind])
   return tree.actionArgPair()

def inWhitelist(env, pos, delta, whitelist):
   r, c = posSum(pos, delta)
   return env[r, c] in whitelist

def whitelistByBlock(delMatPairs, whitelist):
   ret = []
   for deli, mati in delMatPairs:
      if mati in whitelist:
         ret += [deli]
   return ret

def l1(pos1, pos2):
   r1, c1 = pos1
   r2, c2 = pos2
   return abs(r1-r2) + abs(c1-c2)

def sortedByl1(ent, sz, startEnt):
   targs = l1Range(ent, sz, startEnt.pos, startEnt.searchRange)
   targs = sorted(targs, key=lambda targ: l1(startEnt.pos, targ.pos))
   return targs

def l1Range(ent, sz, start, rng):
   R, C = sz, sz
   rs, cs = start
   rt = max(0, rs-rng)
   rb = min(R, rs+rng)
   cl = max(0, cs-rng)
   cr = min(C, cs+rng)
   ret = []
   for r in range(rt, rb):
      for c in range(cl, cr):
         if len(ent[r, c]) > 0:
            ret += ent[r, c]
   return ret

def isAdjacent(pos1, pos2):
   rs, re = pos1
   es, ee = pos2
   return np.logical_xor(abs(re - rs) == 1, abs(ee - es) == 1) 

def posSum(pos1, pos2):
   return pos1[0] + pos2[0], pos1[1] + pos2[1]

def routePath(start, end):
   sr, sc = start
   er, ec = end
   if abs(sc - ec) > abs(sr - er):
      return (0, np.sign(ec - sc))
   return (np.sign(er - sr), 0)

 
def inRange(env, start, targ, rng):
   R, C = env.shape
   rs, cs = start
   rt = max(0, rs-rng)
   rb = min(R, rs+rng)
   cl = max(0, cs-rng)
   cr = min(C, cs+rng)
   return targ in env[rt:rb, cl:cr]

#Fix this
def findNearest(env, start, targ, rng=4 ):
   #Quick check
   rs, ts = start
   if not inRange(env, start, targ, rng):
      return

   #Expensive search
   cur = Queue()
   visited = {}
   cur.push(start)
   while not cur.empty:
      r, c = cur.pop()
      if (r, c) in visited:
         continue

      if env[r, c] == targ:
         return (r, c)

      visited[(r, c)] = 1
      if rs - r < targ:
         cur.push(r+1,  c)
      if cs - c < targ:
         cur.push(r,  c+1)
      if r - rs < targ:
         cur.push(r-1, c) 
      if c - cs < targ:
         cur.push(r, c-1)

   return None
   
class RageClock:
   def __init__(self, ticks):
      self.ticks = ticks

   def tick(self):
      self.ticks -= 1

   def isActive(self):
      return self.ticks > 0




from pdb import set_trace as T

from functools import partial
from threading import Thread
from random import random
import time

from bokeh.models import ColumnDataSource
# from bokeh.models.widgets import CheckboxGroup
from bokeh.plotting import curdoc, figure, show

from tornado import gen


# to run a demo with dummy data:
# $ bokeh serve --show visualizer.py

class MarketVisualizer:
    """
    Market Visualizer
    Visualizes a stream of data, automatically refreshes on update()
    """


    def __init__(self, keys: list, history_len: int = 10,
                 title: str = "NeuralMMO Market Data", x: str = "tick",
                 ylabel: str = "Dummy Values"):
        """
        Args:
            keys (list):       List of object names (str) to be displayed on
                               the market
            history_len (int): How far back to plot data, default is 10 ticks
            title (str):       Title of graph
            x (str):           Name of x value, both in data representation and
                               x axis
            ylabel (str):      Name of y axis on plot
            seed (int):        seed for random number generation
        """
        COLORS = 'blue red green yellow black purple'.split()

        self.history_len = history_len
        self.title       = title
        self.keys        = keys

        self.ylabel      = ylabel
        self.x           = x

        self.data        = {x: [0]}
        self.colors      = {}

        # TODO figure out a better way to ensure each 
        # object has a unique color
        assert len(keys) <= len(COLORS), 'Limited color pool'

        for i, key in enumerate(keys):
            self.data[key] = [0.5]
            self.colors[key] = COLORS[i]

        # this must only be modified from a Bokeh session callback
        self.source = ColumnDataSource(data=self.data)

        # This is important! Save curdoc() to make sure all threads
        # see the same document.
        self.doc = curdoc()
        self.p = figure(
           title='Neural MMO: Market Data',
           x_axis_label=x,
           y_axis_label=ylabel)

        for key in keys:
            self.p.line(x=x, y=key, source=self.source,
                        color=self.colors[key], legend_label=key)

        # Hide lines that you click on in legend,
        # ease of use for visualization
        self.p.legend.click_policy = "hide"
        self.doc.add_root(self.p)
        show(self.p)

        # market plans
        # click on objects in market to display stats about them
        # market overview tab
        # -> trade bandwidth, demand, gdp

    @gen.coroutine
    def stream(self):
        # stream new data to document
        self.source.stream(self.data, self.history_len)

    def update(self):
        """
        Update fuction to be used externally.
        data (dict): updated data to be added to market, including tick value
        """
        # Update the document from callback
        self.doc.add_next_tick_callback(partial(self.stream))

def recv_dummy_data(market_visualizer):
    while True:
        # do some blocking computation
        time.sleep(1)

        #Best practice: update in one tick
        data = {}

        for key, val in market_visualizer.data.items():
            data[key] = val
            # increments tick (x axis)
            if key == mv.x:
                data[key].append(val[-1] + 1)
            # adds r in [-0.5, 0.5) to each value
            else:
                data[key].append(val[-1] + random() - 0.5)
        mv.data = data
        mv.update()

# Example setup
#MARKET_ITEMS = ['Food', 'Water', 'Sword']
#mv = MarketVisualizer(MARKET_ITEMS)

#thread = Thread(target=recv_dummy_data, args=[mv])
#thread.start()



from pdb import set_trace as T
import numpy as np

import torch
from torch import nn
from torch.nn import functional as F

from neural_mmo.forge.blade.io import node

class Embedding(nn.Module):
   def __init__(self, var, dim):
      '''Pytorch embedding wrapper that subtracts the min'''
      super().__init__()
      self.embed = torch.nn.Embedding(var.range, dim)
      self.min = var.min

   def forward(self, x):
      return self.embed(x - self.min)

class Input(nn.Module):
   def __init__(self, cls, config):
      '''Embedding wrapper around discrete and continuous vals'''
      super().__init__()
      self.cls = cls
      if isinstance(cls, node.Discrete):
         self.embed = Embedding(cls, config.EMBED)
      elif isinstance(cls, node.Continuous):
         self.embed = torch.nn.Linear(1, config.EMBED)

   def forward(self, x):
      if isinstance(self.cls, node.Discrete):
         x = x.long()
      elif isinstance(self.cls, node.Continuous):
         x = x.float().unsqueeze(-1)
      x = self.embed(x)
      return x

class BiasedInput(nn.Module):
   def __init__(self, cls, config):
      '''Adds a bias to nn.Embedding
      This is useful for attentional models
      to learn a sort of positional embedding'''
      super().__init__()
      self.bias  = torch.nn.Embedding(1, config.HIDDEN)
      self.embed = Input(cls, config)

   def forward(self, x):
      return self.embed(x) + self.bias.weight

class MixedDTypeInput(nn.Module):
   def __init__(self, continuous, discrete, config):
      super().__init__()

      self.continuous = torch.nn.ModuleList([
            torch.nn.Linear(1, config.EMBED) for _ in range(continuous)])
      self.discrete   = torch.nn.Embedding(discrete, config.EMBED)

   def forward(self, x):
      continuous = x['Continuous'].split(1, dim=-1)
      continuous = [net(e) for net, e in zip(self.continuous, continuous)]
      continuous = torch.stack(continuous, dim=-2)
      discrete   = self.discrete(x['Discrete'].long())

      return torch.cat((continuous, discrete), dim=-2)



from neural_mmo.forge.blade import lib

class Recipe:
   def __init__(self, *args, amtMade=1):
      self.amtMade = amtMade 
      self.blueprint = lib.MultiSet()
      for i in range(0, len(args), 2):
         inp = args[i]
         amt = args[i+1]
         self.blueprint.add(inp, amt)
